@@ -2,7 +2,7 @@
 
   cont.c - 
 
-  $Author: akr $
+  $Author: yugui $
   created at: Thu May 23 09:03:43 2007
 
   Copyright (C) 2007 Koichi Sasada
@@ -14,16 +14,24 @@
 #include "gc.h"
 #include "eval_intern.h"
 
+#define CAPTURE_JUST_VALID_VM_STACK 1
+
 enum context_type {
     CONTINUATION_CONTEXT = 0,
     FIBER_CONTEXT = 1,
-    ROOT_FIBER_CONTEXT = 2,
+    ROOT_FIBER_CONTEXT = 2
 };
 
 typedef struct rb_context_struct {
+    enum context_type type;
     VALUE self;
+    int argc;
     VALUE value;
     VALUE *vm_stack;
+#ifdef CAPTURE_JUST_VALID_VM_STACK
+    int vm_stack_slen;  /* length of stack (head of th->stack) */
+    int vm_stack_clen;  /* length of control frames (tail of th->stack) */
+#endif
     VALUE *machine_stack;
     VALUE *machine_stack_src;
 #ifdef __ia64
@@ -34,11 +42,22 @@ typedef struct rb_context_struct {
     rb_thread_t saved_thread;
     rb_jmpbuf_t jmpbuf;
     int machine_stack_size;
-    VALUE prev;
-    int alive;
-    enum context_type type;
 } rb_context_t;
 
+enum fiber_status {
+    CREATED,
+    RUNNING,
+    TERMINATED
+};
+
+typedef struct rb_fiber_struct {
+    rb_context_t cont;
+    VALUE prev;
+    enum fiber_status status;
+    struct rb_fiber_struct *prev_fiber;
+    struct rb_fiber_struct *next_fiber;
+} rb_fiber_t;
+
 static VALUE rb_cContinuation;
 static VALUE rb_cFiber;
 static VALUE rb_eFiberError;
@@ -46,6 +65,11 @@ static VALUE rb_eFiberError;
 #define GetContPtr(obj, ptr)  \
   Data_Get_Struct(obj, rb_context_t, ptr)
 
+#define GetFiberPtr(obj, ptr)  do {\
+  ptr = (rb_fiber_t*)DATA_PTR(obj);\
+  if (!ptr) rb_raise(rb_eFiberError, "uninitialized fiber");\
+} while(0)
+
 NOINLINE(static VALUE cont_capture(volatile int *stat));
 
 void rb_thread_mark(rb_thread_t *th);
@@ -57,12 +81,16 @@ cont_mark(void *ptr)
     if (ptr) {
 	rb_context_t *cont = ptr;
 	rb_gc_mark(cont->value);
-	rb_gc_mark(cont->prev);
 	rb_thread_mark(&cont->saved_thread);
 
 	if (cont->vm_stack) {
+#ifdef CAPTURE_JUST_VALID_VM_STACK
 	    rb_gc_mark_locations(cont->vm_stack,
-				 cont->vm_stack + cont->saved_thread.stack_size);
+				 cont->vm_stack + cont->vm_stack_slen + cont->vm_stack_clen);
+#else
+	    rb_gc_mark_localtion(cont->vm_stack,
+				 cont->vm_stack, cont->saved_thread.stack_size);
+#endif
 	}
 
 	if (cont->machine_stack) {
@@ -85,22 +113,69 @@ cont_free(void *ptr)
     RUBY_FREE_ENTER("cont");
     if (ptr) {
 	rb_context_t *cont = ptr;
-	RUBY_FREE_UNLESS_NULL(cont->saved_thread.stack);
+	RUBY_FREE_UNLESS_NULL(cont->saved_thread.stack); fflush(stdout);
 	RUBY_FREE_UNLESS_NULL(cont->machine_stack);
 #ifdef __ia64
 	RUBY_FREE_UNLESS_NULL(cont->machine_register_stack);
 #endif
 	RUBY_FREE_UNLESS_NULL(cont->vm_stack);
 
-	if (cont->type == FIBER_CONTEXT) {
-	    st_free_table(cont->saved_thread.local_storage);
-	}
-
+	/* free rb_cont_t or rb_fiber_t */
 	ruby_xfree(ptr);
     }
     RUBY_FREE_LEAVE("cont");
 }
 
+static void
+fiber_mark(void *ptr)
+{
+    RUBY_MARK_ENTER("cont");
+    if (ptr) {
+	rb_fiber_t *fib = ptr;
+	rb_gc_mark(fib->prev);
+	cont_mark(&fib->cont);
+    }
+    RUBY_MARK_LEAVE("cont");
+}
+
+static void
+fiber_link_join(rb_fiber_t *fib)
+{
+    VALUE current_fibval = rb_fiber_current();
+    rb_fiber_t *current_fib;
+    GetFiberPtr(current_fibval, current_fib);
+
+    /* join fiber link */
+    fib->next_fiber = current_fib->next_fiber;
+    fib->prev_fiber = current_fib;
+    current_fib->next_fiber->prev_fiber = fib;
+    current_fib->next_fiber = fib;
+}
+
+static void
+fiber_link_remove(rb_fiber_t *fib)
+{
+    fib->prev_fiber->next_fiber = fib->next_fiber;
+    fib->next_fiber->prev_fiber = fib->prev_fiber;
+}
+
+static void
+fiber_free(void *ptr)
+{
+    RUBY_FREE_ENTER("fiber");
+    if (ptr) {
+	rb_fiber_t *fib = ptr;
+
+	if (fib->cont.type != ROOT_FIBER_CONTEXT) {
+	    st_free_table(fib->cont.saved_thread.local_storage);
+	}
+	fiber_link_remove(fib);
+
+	cont_free(&fib->cont);
+    }
+    RUBY_FREE_LEAVE("fiber");
+}
+
 static void
 cont_save_machine_stack(rb_thread_t *th, rb_context_t *cont)
 {
@@ -151,26 +226,28 @@ cont_save_machine_stack(rb_thread_t *th, rb_context_t *cont)
 #endif
 }
 
+static void
+cont_init(rb_context_t *cont)
+{
+    rb_thread_t *th = GET_THREAD();
+
+    /* save thread context */
+    cont->saved_thread = *th;
+}
+
 static rb_context_t *
 cont_new(VALUE klass)
 {
     rb_context_t *cont;
     volatile VALUE contval;
-    rb_thread_t *th = GET_THREAD();
-
-    contval = Data_Make_Struct(klass, rb_context_t,
-			       cont_mark, cont_free, cont);
 
+    contval = Data_Make_Struct(klass, rb_context_t, cont_mark, cont_free, cont);
     cont->self = contval;
-    cont->alive = Qtrue;
-
-    /* save thread context */
-    cont->saved_thread = *th;
-
+    cont_init(cont);
     return cont;
 }
 
-void vm_stack_to_heap(rb_thread_t *th);
+void rb_vm_stack_to_heap(rb_thread_t *th);
 
 static VALUE
 cont_capture(volatile int *stat)
@@ -179,13 +256,21 @@ cont_capture(volatile int *stat)
     rb_thread_t *th = GET_THREAD(), *sth;
     volatile VALUE contval;
 
-    vm_stack_to_heap(th);
+    rb_vm_stack_to_heap(th);
     cont = cont_new(rb_cContinuation);
     contval = cont->self;
     sth = &cont->saved_thread;
 
+#ifdef CAPTURE_JUST_VALID_VM_STACK
+    cont->vm_stack_slen = th->cfp->sp + th->mark_stack_len - th->stack;
+    cont->vm_stack_clen = th->stack + th->stack_size - (VALUE*)th->cfp;
+    cont->vm_stack = ALLOC_N(VALUE, cont->vm_stack_slen + cont->vm_stack_clen);
+    MEMCPY(cont->vm_stack, th->stack, VALUE, cont->vm_stack_slen);
+    MEMCPY(cont->vm_stack + cont->vm_stack_slen, (VALUE*)th->cfp, VALUE, cont->vm_stack_clen);
+#else
     cont->vm_stack = ALLOC_N(VALUE, th->stack_size);
     MEMCPY(cont->vm_stack, th->stack, VALUE, th->stack_size);
+#endif
     sth->stack = 0;
 
     cont_save_machine_stack(th, cont);
@@ -204,7 +289,7 @@ cont_capture(volatile int *stat)
     }
 }
 
-NORETURN(static void cont_restore_1(rb_context_t *));
+NOINLINE(NORETURN(static void cont_restore_1(rb_context_t *)));
 
 static void
 cont_restore_1(rb_context_t *cont)
@@ -225,7 +310,13 @@ cont_restore_1(rb_context_t *cont)
 	    th->stack_size = fcont->saved_thread.stack_size;
 	    th->stack = fcont->saved_thread.stack;
 	}
+#ifdef CAPTURE_JUST_VALID_VM_STACK
+	MEMCPY(th->stack, cont->vm_stack, VALUE, cont->vm_stack_slen);
+	MEMCPY(th->stack + sth->stack_size - cont->vm_stack_clen,
+	       cont->vm_stack + cont->vm_stack_slen, VALUE, cont->vm_stack_clen);
+#else
 	MEMCPY(th->stack, cont->vm_stack, VALUE, sth->stack_size);
+#endif
     }
     else {
 	/* fiber */
@@ -323,7 +414,7 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
 #else
 	if (addr_in_prev_frame > &space[0]) {
 	    /* Stack grows downward */
-	    if (addr_in_prev_frame > cont->saved_thread.machine_stack_src) {
+	    if (addr_in_prev_frame > cont->machine_stack_src) {
 		cont_restore_0(cont, &space[0]);
 	    }
 	}
@@ -466,12 +557,9 @@ rb_cont_call(int argc, VALUE *argv, VALUE contval)
 	if (th->fiber != cont->saved_thread.fiber) {
 	    rb_raise(rb_eRuntimeError, "continuation called across fiber");
 	}
-	
-	if (!fcont->alive) {
-	    rb_raise(rb_eRuntimeError, "continuation called dead fiber");
-	}
     }
 
+    cont->argc = argc;
     cont->value = make_passing_arg(argc, argv);
 
     cont_restore_0(cont, &contval);
@@ -482,27 +570,103 @@ rb_cont_call(int argc, VALUE *argv, VALUE contval)
 /* fiber */
 /*********/
 
+/*
+ *  Document-class: Fiber
+ *
+ *  Fibers are primitives for implementing light weight cooperative
+ *  concurrency in Ruby. Basically they are a means of creating code blocks 
+ *  that can be paused and resumed, much like threads. The main difference 
+ *  is that they are never preempted and that the scheduling must be done by 
+ *  the programmer and not the VM. 
+ *
+ *  As opposed to other stackless light weight concurrency models, each fiber
+ *  comes with a small 4KB stack. This enables the fiber to be paused from deeply
+ *  nested function calls within the fiber block.
+ *
+ *  When a fiber is created it will not run automatically. Rather it must be 
+ *  be explicitly asked to run using the <code>Fiber#resume</code> method. 
+ *  The code running inside the fiber can give up control by calling 
+ *  <code>Fiber.yield</code> in which case it yields control back to caller 
+ *  (the caller of the <code>Fiber#resume</code>).
+ * 
+ *  Upon yielding or termination the Fiber returns the value of the last 
+ *  executed expression
+ *  
+ *  For instance:
+ *  
+ *    fiber = Fiber.new do
+ *      Fiber.yield 1
+ *      2
+ *    end
+ *
+ *    puts fiber.resume
+ *    puts fiber.resume
+ *    puts fiber.resume
+ *    
+ *  <em>produces</em>
+ *    
+ *    1
+ *    2
+ *    FiberError: dead fiber called
+ *     
+ *  The <code>Fiber#resume</code> method accepts an arbitary number of
+ *  parameters, if it is the first call to <code>resume</code> then they
+ *  will be passed as block arguments. Otherwise they will be the return
+ *  value of the call to <code>Fiber.yield</code>
+ *
+ *  Example:
+ *  
+ *    fiber = Fiber.new do |first|
+ *      second = Fiber.yield first + 2
+ *    end
+ *
+ *    puts fiber.resume 10
+ *    puts fiber.resume 14
+ *    puts fiber.resume 18
+ *
+ *  <em>produces</em>
+ *    
+ *    12
+ *    14
+ *    FiberError: dead fiber called
+ *
+ */
+
 #define FIBER_VM_STACK_SIZE (4 * 1024)
 
-static rb_context_t *
+static VALUE
 fiber_alloc(VALUE klass)
 {
-    rb_context_t *cont = cont_new(klass);
+    return Data_Wrap_Struct(klass, fiber_mark, fiber_free, 0);
+}
 
-    cont->type = FIBER_CONTEXT;
-    cont->prev = Qnil;
+static rb_fiber_t*
+fiber_t_alloc(VALUE fibval)
+{
+    rb_fiber_t *fib = ALLOC(rb_fiber_t);
 
-    return cont;
+    memset(fib, 0, sizeof(rb_fiber_t));
+    fib->cont.self = fibval;
+    fib->cont.type = FIBER_CONTEXT;
+    cont_init(&fib->cont);
+    fib->prev = Qnil;
+    fib->status = CREATED;
+
+    DATA_PTR(fibval) = fib;
+
+    return fib;
 }
 
 static VALUE
-fiber_new(VALUE klass, VALUE proc)
+fiber_init(VALUE fibval, VALUE proc)
 {
-    rb_context_t *cont = fiber_alloc(klass);
-    VALUE contval = cont->self;
+    rb_fiber_t *fib = fiber_t_alloc(fibval);
+    rb_context_t *cont = &fib->cont;
     rb_thread_t *th = &cont->saved_thread;
 
-    /* initialize */
+    fiber_link_join(fib);
+
+    /* initialize cont */
     cont->vm_stack = 0;
 
     th->stack = 0;
@@ -529,29 +693,29 @@ fiber_new(VALUE klass, VALUE proc)
 
     MEMCPY(&cont->jmpbuf, &th->root_jmpbuf, rb_jmpbuf_t, 1);
 
-    return contval;
+    return fibval;
+}
+
+static VALUE
+rb_fiber_init(VALUE fibval)
+{
+    return fiber_init(fibval, rb_block_proc());
 }
 
 VALUE
 rb_fiber_new(VALUE (*func)(ANYARGS), VALUE obj)
 {
-    return fiber_new(rb_cFiber, rb_proc_new(func, obj));
-}
-
-static VALUE
-rb_fiber_s_new(VALUE self)
-{
-    return fiber_new(self, rb_block_proc());
+    return fiber_init(fiber_alloc(rb_cFiber), rb_proc_new(func, obj));
 }
 
 static VALUE
 return_fiber(void)
 {
-    rb_context_t *cont;
+    rb_fiber_t *fib;
     VALUE curr = rb_fiber_current();
-    GetContPtr(curr, cont);
+    GetFiberPtr(curr, fib);
 
-    if (cont->prev == Qnil) {
+    if (fib->prev == Qnil) {
 	rb_thread_t *th = GET_THREAD();
 
 	if (th->root_fiber != curr) {
@@ -562,8 +726,8 @@ return_fiber(void)
 	}
     }
     else {
-	VALUE prev = cont->prev;
-	cont->prev = Qnil;
+	VALUE prev = fib->prev;
+	fib->prev = Qnil;
 	return prev;
     }
 }
@@ -571,10 +735,10 @@ return_fiber(void)
 VALUE rb_fiber_transfer(VALUE fib, int argc, VALUE *argv);
 
 static void
-rb_fiber_terminate(rb_context_t *cont)
+rb_fiber_terminate(rb_fiber_t *fib)
 {
-    VALUE value = cont->value;
-    cont->alive = Qfalse;
+    VALUE value = fib->cont.value;
+    fib->status = TERMINATED;
     rb_fiber_transfer(return_fiber(), 1, &value);
 }
 
@@ -582,22 +746,28 @@ void
 rb_fiber_start(void)
 {
     rb_thread_t *th = GET_THREAD();
+    rb_fiber_t *fib;
     rb_context_t *cont;
     rb_proc_t *proc;
-    VALUE args;
     int state;
 
-    GetContPtr(th->fiber, cont);
+    GetFiberPtr(th->fiber, fib);
+    cont = &fib->cont;
+
     TH_PUSH_TAG(th);
     if ((state = EXEC_TAG()) == 0) {
+	int argc;
+	VALUE *argv, args;
 	GetProcPtr(cont->saved_thread.first_proc, proc);
 	args = cont->value;
+	argv = (argc = cont->argc) > 1 ? RARRAY_PTR(args) : &args;
 	cont->value = Qnil;
 	th->errinfo = Qnil;
 	th->local_lfp = proc->block.lfp;
 	th->local_svar = Qnil;
 
-	cont->value = vm_invoke_proc(th, proc, proc->block.self, 1, &args, 0);
+	fib->status = RUNNING;
+	cont->value = rb_vm_invoke_proc(th, proc, proc->block.self, argc, argv, 0);
     }
     TH_POP_TAG();
 
@@ -607,51 +777,62 @@ rb_fiber_start(void)
 	}
 	else {
 	    th->thrown_errinfo =
-	      vm_make_jump_tag_but_local_jump(state, th->errinfo);
+	      rb_vm_make_jump_tag_but_local_jump(state, th->errinfo);
 	}
 	RUBY_VM_SET_INTERRUPT(th);
     }
 
-    rb_fiber_terminate(cont);
+    rb_fiber_terminate(fib);
     rb_bug("rb_fiber_start: unreachable");
 }
 
+static rb_fiber_t *
+root_fiber_alloc(rb_thread_t *th)
+{
+    rb_fiber_t *fib;
+
+    /* no need to allocate vm stack */
+    fib = fiber_t_alloc(fiber_alloc(rb_cFiber));
+    fib->cont.type = ROOT_FIBER_CONTEXT;
+    fib->prev_fiber = fib->next_fiber = fib;
+
+    return fib;
+}
+
 VALUE
 rb_fiber_current()
 {
     rb_thread_t *th = GET_THREAD();
     if (th->fiber == 0) {
 	/* save root */
-	rb_context_t *cont = fiber_alloc(rb_cFiber);
-	cont->type = ROOT_FIBER_CONTEXT;
-	th->root_fiber = th->fiber = cont->self;
+	rb_fiber_t *fib = root_fiber_alloc(th);
+	th->root_fiber = th->fiber = fib->cont.self;
     }
     return th->fiber;
 }
 
 static VALUE
-fiber_store(rb_context_t *next_cont)
+fiber_store(rb_fiber_t *next_fib)
 {
     rb_thread_t *th = GET_THREAD();
-    rb_context_t *cont;
+    rb_fiber_t *fib;
 
     if (th->fiber) {
-	GetContPtr(th->fiber, cont);
-	cont->saved_thread = *th;
+	GetFiberPtr(th->fiber, fib);
+	fib->cont.saved_thread = *th;
     }
     else {
 	/* create current fiber */
-	cont = fiber_alloc(rb_cFiber); /* no need to allocate vm stack */
-	cont->type = ROOT_FIBER_CONTEXT;
-	th->root_fiber = th->fiber = cont->self;
+	fib = root_fiber_alloc(th);
+	th->root_fiber = th->fiber = fib->cont.self;
     }
 
-    cont_save_machine_stack(th, cont);
+    cont_save_machine_stack(th, &fib->cont);
 
-    if (ruby_setjmp(cont->jmpbuf)) {
+    if (ruby_setjmp(fib->cont.jmpbuf)) {
 	/* restored */
-	GetContPtr(th->fiber, cont);
-	return cont->value;
+	GetFiberPtr(th->fiber, fib);
+	return fib->cont.value;
     }
     else {
 	return Qundef;
@@ -659,13 +840,15 @@ fiber_store(rb_context_t *next_cont)
 }
 
 static inline VALUE
-fiber_switch(VALUE fib, int argc, VALUE *argv, int is_resume)
+fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
 {
     VALUE value;
+    rb_fiber_t *fib;
     rb_context_t *cont;
     rb_thread_t *th = GET_THREAD();
 
-    GetContPtr(fib, cont);
+    GetFiberPtr(fibval, fib);
+    cont = &fib->cont;
 
     if (cont->saved_thread.self != th->self) {
 	rb_raise(rb_eFiberError, "fiber called across threads");
@@ -673,18 +856,19 @@ fiber_switch(VALUE fib, int argc, VALUE *argv, int is_resume)
     else if (cont->saved_thread.trap_tag != th->trap_tag) {
 	rb_raise(rb_eFiberError, "fiber called across trap");
     }
-    else if (!cont->alive) {
+    else if (fib->status == TERMINATED) {
 	rb_raise(rb_eFiberError, "dead fiber called");
     }
 
     if (is_resume) {
-	cont->prev = rb_fiber_current();
+	fib->prev = rb_fiber_current();
     }
 
+    cont->argc = argc;
     cont->value = make_passing_arg(argc, argv);
 
-    if ((value = fiber_store(cont)) == Qundef) {
-	cont_restore_0(cont, &value);
+    if ((value = fiber_store(fib)) == Qundef) {
+	cont_restore_0(&fib->cont, &value);
 	rb_bug("rb_fiber_resume: unreachable");
     }
 
@@ -700,16 +884,16 @@ rb_fiber_transfer(VALUE fib, int argc, VALUE *argv)
 }
 
 VALUE
-rb_fiber_resume(VALUE fib, int argc, VALUE *argv)
+rb_fiber_resume(VALUE fibval, int argc, VALUE *argv)
 {
-    rb_context_t *cont;
-    GetContPtr(fib, cont);
+    rb_fiber_t *fib;
+    GetFiberPtr(fibval, fib);
 
-    if (cont->prev != Qnil) {
+    if (fib->prev != Qnil) {
 	rb_raise(rb_eFiberError, "double resume");
     }
 
-    return fiber_switch(fib, argc, argv, 1);
+    return fiber_switch(fibval, argc, argv, 1);
 }
 
 VALUE
@@ -718,32 +902,89 @@ rb_fiber_yield(int argc, VALUE *argv)
     return rb_fiber_transfer(return_fiber(), argc, argv);
 }
 
+/*
+ *  call-seq:
+ *     fiber.alive? -> true or false
+ *  
+ *  Returns true if the fiber can still be resumed (or transferred to).
+ *  After finishing execution of the fiber block this method will always
+ *  return false.
+ */
 VALUE
-rb_fiber_alive_p(VALUE fib)
+rb_fiber_alive_p(VALUE fibval)
 {
-    rb_context_t *cont;
-    GetContPtr(fib, cont);
-    return cont->alive;
+    rb_fiber_t *fib;
+    GetFiberPtr(fibval, fib);
+    return fib->status != TERMINATED;
 }
 
+/*
+ *  call-seq:
+ *     fiber.resume(args, ...) -> obj
+ *  
+ *  Resumes the fiber from the point at which the last <code>Fiber.yield</code>
+ *  was called, or starts running it if it is the first call to 
+ *  <code>resume</code>. Arguments passed to resume will be the value of
+ *  the <code>Fiber.yield</code> expression or will be passed as block 
+ *  parameters to the fiber's block if this is the first <code>resume</code>.
+ *  
+ *  Alternatively, when resume is called it evaluates to the arguments passed
+ *  to the next <code>Fiber.yield</code> statement inside the fiber's block
+ *  or to the block value if it runs to completion without any
+ *  <code>Fiber.yield</code>
+ */
 static VALUE
 rb_fiber_m_resume(int argc, VALUE *argv, VALUE fib)
 {
     return rb_fiber_resume(fib, argc, argv);
 }
 
+/*
+ *  call-seq:
+ *     fiber.transfer(args, ...) -> obj
+ *  
+ *  Transfer control to another fiber, resuming it from where it last
+ *  stopped or starting it if it was not resumed before. The calling 
+ *  fiber will be suspended much like in a call to <code>Fiber.yield</code>.
+ *  
+ *  The fiber which recieves the transfer call is treats it much like 
+ *  a resume call. Arguments passed to transfer are treated like those
+ *  passed to resume.
+ *     
+ *  You cannot resume a fiber that transferred control to another one.
+ *  This will cause a double resume error. You need to transfer control
+ *  back to this fiber before it can yield and resume.
+ */
 static VALUE
 rb_fiber_m_transfer(int argc, VALUE *argv, VALUE fib)
 {
     return rb_fiber_transfer(fib, argc, argv);
 }
 
+/*
+ *  call-seq:
+ *     Fiber.yield(args, ...) -> obj
+ *  
+ *  Yields control back to the context that resumed the fiber, passing
+ *  along any arguments that were passed to it. The fiber will resume
+ *  processing at this point when <code>resume</code> is called next.
+ *  Any arguments passed to the next <code>resume</code> will be the
+ *  value that this <code>Fiber.yield</code> expression evaluates to.
+ */
 static VALUE
 rb_fiber_s_yield(int argc, VALUE *argv, VALUE klass)
 {
     return rb_fiber_yield(argc, argv);
 }
 
+/*
+ *  call-seq:
+ *     Fiber.current() -> fiber
+ *  
+ *  Returns the current fiber. You need to <code>require 'fiber'</code>
+ *  before using this method. If you are not running in the context of
+ *  a fiber this method will return the root fiber.
+ */
 static VALUE
 rb_fiber_s_current(VALUE klass)
 {
@@ -754,15 +995,15 @@ void
 Init_Cont(void)
 {
     rb_cFiber = rb_define_class("Fiber", rb_cObject);
-    rb_undef_alloc_func(rb_cFiber);
+    rb_define_alloc_func(rb_cFiber, fiber_alloc);
     rb_eFiberError = rb_define_class("FiberError", rb_eStandardError);
-    rb_define_singleton_method(rb_cFiber, "new", rb_fiber_s_new, 0);
     rb_define_singleton_method(rb_cFiber, "yield", rb_fiber_s_yield, -1);
+    rb_define_method(rb_cFiber, "initialize", rb_fiber_init, 0);
     rb_define_method(rb_cFiber, "resume", rb_fiber_m_resume, -1);
 }
 
 void
-Init_Continuation_body(void)
+ruby_Init_Continuation_body(void)
 {
     rb_cContinuation = rb_define_class("Continuation", rb_cObject);
     rb_undef_alloc_func(rb_cContinuation);
@@ -773,7 +1014,7 @@ Init_Continuation_body(void)
 }
 
 void
-Init_Fiber_as_Coroutine(void)
+ruby_Init_Fiber_as_Coroutine(void)
 {
     rb_define_method(rb_cFiber, "transfer", rb_fiber_m_transfer, -1);
     rb_define_method(rb_cFiber, "alive?", rb_fiber_alive_p, 0);
