@@ -2,7 +2,7 @@
 
   dir.c -
 
-  $Author: naruse $
+  $Author: yugui $
   created at: Wed Jan  5 09:51:01 JST 1994
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -87,7 +87,7 @@ static int
 char_casecmp(const char *p1, const char *p2, rb_encoding *enc, const int nocase)
 {
     const char *p1end, *p2end;
-    int c1, c2;
+    unsigned int c1, c2;
 
     if (!*p1 || !*p2) return !!*p1 - !!*p2;
     p1end = p1 + strlen(p1);
@@ -290,7 +290,7 @@ VALUE rb_cDir;
 struct dir_data {
     DIR *dir;
     VALUE path;
-    rb_encoding *extenc;
+    rb_encoding *enc;
 };
 
 static void
@@ -318,7 +318,7 @@ dir_s_alloc(VALUE klass)
 
     dirp->dir = NULL;
     dirp->path = Qnil;
-    dirp->extenc = NULL;
+    dirp->enc = NULL;
 
     return obj;
 }
@@ -333,26 +333,26 @@ static VALUE
 dir_initialize(int argc, VALUE *argv, VALUE dir)
 {
     struct dir_data *dp;
-    rb_encoding  *extencoding;
+    rb_encoding  *fsenc;
     VALUE dirname, opt;
-    static VALUE sym_extenc;
+    static VALUE sym_enc;
 
-    if (!sym_extenc) {
-	sym_extenc = ID2SYM(rb_intern("external_encoding"));
+    if (!sym_enc) {
+	sym_enc = ID2SYM(rb_intern("encoding"));
     }
-    extencoding = rb_filesystem_encoding();
+    fsenc = rb_filesystem_encoding();
 
     rb_scan_args(argc, argv, "11", &dirname, &opt);
 
     if (!NIL_P(opt)) {
-        VALUE v, extenc=Qnil;
+        VALUE v, enc=Qnil;
         opt = rb_convert_type(opt, T_HASH, "Hash", "to_hash");
 
-        v = rb_hash_aref(opt, sym_extenc);
-        if (!NIL_P(v)) extenc = v;
+        v = rb_hash_aref(opt, sym_enc);
+        if (!NIL_P(v)) enc = v;
 
-	if (!NIL_P(extenc)) {
-	    extencoding = rb_to_encoding(extenc);
+	if (!NIL_P(enc)) {
+	    fsenc = rb_to_encoding(enc);
 	}
     }
 
@@ -362,7 +362,7 @@ dir_initialize(int argc, VALUE *argv, VALUE dir)
     if (dp->dir) closedir(dp->dir);
     dp->dir = NULL;
     dp->path = Qnil;
-    dp->extenc = extencoding;
+    dp->enc = fsenc;
     dp->dir = opendir(RSTRING_PTR(dirname));
     if (dp->dir == NULL) {
 	if (errno == EMFILE || errno == ENFILE) {
@@ -423,16 +423,6 @@ dir_check(VALUE dir)
     if (dirp->dir == NULL) dir_closed();\
 } while (0)
 
-static VALUE
-dir_enc_str_new(const char *p, long len, rb_encoding *enc)
-{
-    VALUE path = rb_tainted_str_new(p, len);
-    if (rb_enc_asciicompat(enc) && rb_enc_str_asciionly_p(path)) {
-	enc = rb_usascii_encoding();
-    }
-    rb_enc_associate(path, enc);
-    return path;
-}
 
 /*
  *  call-seq:
@@ -494,7 +484,7 @@ dir_read(VALUE dir)
     errno = 0;
     dp = readdir(dirp->dir);
     if (dp) {
-	return dir_enc_str_new(dp->d_name, NAMLEN(dp), dirp->extenc);
+	return rb_external_str_new_with_enc(dp->d_name, NAMLEN(dp), dirp->enc);
     }
     else if (errno == 0) {	/* end of stream */
 	return Qnil;
@@ -532,7 +522,7 @@ dir_each(VALUE dir)
     GetDIR(dir, dirp);
     rewinddir(dirp->dir);
     for (dp = readdir(dirp->dir); dp != NULL; dp = readdir(dirp->dir)) {
-	rb_yield(dir_enc_str_new(dp->d_name, NAMLEN(dp), dirp->extenc));
+	rb_yield(rb_external_str_new_with_enc(dp->d_name, NAMLEN(dp), dirp->enc));
 	if (dirp->dir == NULL) dir_closed();
     }
     return dir;
@@ -1122,15 +1112,16 @@ static char *
 join_path(const char *path, int dirsep, const char *name)
 {
     long len = strlen(path);
-    char *buf = GLOB_ALLOC_N(char, len+strlen(name)+(dirsep?1:0)+1);
+    long len2 = strlen(name)+(dirsep?1:0)+1;
+    char *buf = GLOB_ALLOC_N(char, len+len2);
 
     if (!buf) return 0;
     memcpy(buf, path, len);
     if (dirsep) {
-	strcpy(buf+len, "/");
-	len++;
+	buf[len++] = '/';
     }
-    strcpy(buf+len, name);
+    buf[len] = '\0';
+    strlcat(buf+len, name, len2);
     return buf;
 }
 
@@ -1311,12 +1302,13 @@ glob_helper(
 	    if (*cur) {
 		char *buf;
 		char *name;
-		name = GLOB_ALLOC_N(char, strlen((*cur)->str) + 1);
+		size_t len = strlen((*cur)->str) + 1;
+		name = GLOB_ALLOC_N(char, len);
 		if (!name) {
 		    status = -1;
 		    break;
 		}
-		strcpy(name, (*cur)->str);
+		memcpy(name, (*cur)->str, len);
 		if (escape) remove_backslashes(name, enc);
 
 		new_beg = new_end = GLOB_ALLOC_N(struct glob_pattern *, end - beg);
@@ -1436,7 +1428,7 @@ rb_glob(const char *path, void (*func)(const char *, VALUE, void *), VALUE arg)
 static void
 push_pattern(const char *path, VALUE ary, void *enc)
 {
-    rb_ary_push(ary, dir_enc_str_new(path, strlen(path), enc));
+    rb_ary_push(ary, rb_external_str_new_with_enc(path, strlen(path), enc));
 }
 
 static int
@@ -1581,7 +1573,7 @@ dir_globs(long argc, VALUE *argv, int flags)
     for (i = 0; i < argc; ++i) {
 	int status;
 	VALUE str = argv[i];
-	StringValue(str);
+	SafeStringValue(str);
 	status = push_glob(ary, str, flags);
 	if (status) GLOB_JUMP_TAG(status);
     }
