@@ -1,5 +1,5 @@
 /*
- * $Id: ossl_config.c 28367 2010-06-21 09:18:59Z shyouhei $
+ * $Id: ossl_config.c 25945 2009-11-27 02:54:46Z yugui $
  * 'OpenSSL for Ruby' project
  * Copyright (C) 2001-2002  Michal Rokos <m.rokos@sh.cvut.cz>
  * All rights reserved.
@@ -196,21 +196,10 @@ static void
 rb_ossl_config_modify_check(VALUE config)
 {
     if (OBJ_FROZEN(config)) rb_error_frozen("OpenSSL::Config");
-    if (!OBJ_TAINTED(config) && rb_safe_level() >= 4)
+    if (!OBJ_UNTRUSTED(config) && rb_safe_level() >= 4)
 	rb_raise(rb_eSecurityError, "Insecure: can't modify OpenSSL config");
 }
 
-static VALUE
-ossl_config_add_value_m(VALUE self, VALUE section, VALUE name, VALUE value)
-{
-#if defined(OSSL_NO_CONF_API)
-    rb_notimplement();
-#else
-    rb_ossl_config_modify_check(self);
-    return ossl_config_add_value(self, section, name, value);
-#endif
-}
-
 static VALUE
 ossl_config_get_value(VALUE self, VALUE section, VALUE name)
 {
@@ -313,12 +302,6 @@ ossl_config_get_section_old(VALUE self, VALUE section)
 }
 
 #ifdef IMPLEMENT_LHASH_DOALL_ARG_FN
-#define IMPLEMENT_LHASH_DOALL_ARG_FN_098(f_name,o_type,a_type) \
-            void f_name##_LHASH_DOALL_ARG(void *arg1, void *arg2) { \
-		                o_type a = (o_type)arg1; \
-		                a_type b = (a_type)arg2; \
-		                f_name(a,b); }
-
 static void
 get_conf_section(CONF_VALUE *cv, VALUE ary)
 {
@@ -326,7 +309,7 @@ get_conf_section(CONF_VALUE *cv, VALUE ary)
     rb_ary_push(ary, rb_str_new2(cv->section));
 }
 
-static IMPLEMENT_LHASH_DOALL_ARG_FN_098(get_conf_section, CONF_VALUE*, VALUE)
+static IMPLEMENT_LHASH_DOALL_ARG_FN(get_conf_section, CONF_VALUE*, VALUE);
 
 static VALUE
 ossl_config_get_sections(VALUE self)
@@ -364,7 +347,7 @@ dump_conf_value(CONF_VALUE *cv, VALUE str)
     rb_str_cat2(str, "\n");
 }
 
-static IMPLEMENT_LHASH_DOALL_ARG_FN_098(dump_conf_value, CONF_VALUE*, VALUE)
+static IMPLEMENT_LHASH_DOALL_ARG_FN(dump_conf_value, CONF_VALUE*, VALUE);
 
 static VALUE
 dump_conf(CONF *conf)
@@ -408,7 +391,7 @@ each_conf_value(CONF_VALUE *cv, void* dummy)
     }
 }
 
-static IMPLEMENT_LHASH_DOALL_ARG_FN_098(each_conf_value, CONF_VALUE*, void*)
+static IMPLEMENT_LHASH_DOALL_ARG_FN(each_conf_value, CONF_VALUE*, void*);
 
 static VALUE
 ossl_config_each(VALUE self)
@@ -449,7 +432,7 @@ static VALUE
 ossl_config_inspect(VALUE self)
 {
     VALUE str, ary = ossl_config_get_sections(self);
-    char *cname = rb_class2name(rb_obj_class(self));
+    const char *cname = rb_class2name(rb_obj_class(self));
 
     str = rb_str_new2("#<");
     rb_str_cat2(str, cname);
@@ -482,7 +465,7 @@ Init_ossl_config()
     rb_define_method(cConfig, "initialize", ossl_config_initialize, -1);
     rb_define_method(cConfig, "get_value", ossl_config_get_value, 2);
     rb_define_method(cConfig, "value", ossl_config_get_value_old, -1);
-    rb_define_method(cConfig, "add_value", ossl_config_add_value_m, 3);
+    rb_define_method(cConfig, "add_value", ossl_config_add_value, 3);
     rb_define_method(cConfig, "[]", ossl_config_get_section, 1);
     rb_define_method(cConfig, "section", ossl_config_get_section_old, 1);
     rb_define_method(cConfig, "[]=", ossl_config_set_section, 2);
