@@ -1,13 +1,12 @@
 /*
  * node.c
  *
- * $Author: matz $
- * $Date: 2005/09/16 09:35:23 $
+ * $Author: mame $
  *
  * Copyright (C) 2003 why the lucky stiff
  */
 
-#include "ruby.h"
+#include "ruby/ruby.h"
 #include "syck.h"
 
 /*
@@ -46,7 +45,7 @@ syck_free_node( SyckNode *n )
 }
 
 SyckNode *
-syck_alloc_map()
+syck_alloc_map(void)
 {
     SyckNode *n;
     struct SyckMap *m;
@@ -65,7 +64,7 @@ syck_alloc_map()
 }
 
 SyckNode *
-syck_alloc_seq()
+syck_alloc_seq(void)
 {
     SyckNode *n;
     struct SyckSeq *s;
@@ -83,7 +82,7 @@ syck_alloc_seq()
 }
 
 SyckNode *
-syck_alloc_str()
+syck_alloc_str(void)
 {
     SyckNode *n;
     struct SyckStr *s;
@@ -100,13 +99,13 @@ syck_alloc_str()
 }
 
 SyckNode *
-syck_new_str( char *str, enum scalar_style style )
+syck_new_str( const char *str, enum scalar_style style )
 {
     return syck_new_str2( str, strlen( str ), style );
 }
 
 SyckNode *
-syck_new_str2( char *str, long len, enum scalar_style style )
+syck_new_str2( const char *str, long len, enum scalar_style style )
 {
     SyckNode *n;
 
@@ -129,7 +128,7 @@ syck_replace_str( SyckNode *n, char *str, enum scalar_style style )
 void
 syck_replace_str2( SyckNode *n, char *str, long len, enum scalar_style style )
 {
-    if ( n->data.str != NULL ) 
+    if ( n->data.str->ptr != NULL ) 
     {
         S_FREE( n->data.str->ptr );
         n->data.str->ptr = NULL;
