@@ -1,7 +1,6 @@
 #----------------------------------
 # extconf.rb
-# $Revision: 29867 $
-# $Date: 2010-11-22 16:22:26 +0900 (Mon, 22 Nov 2010) $
+# $Revision: 21960 $
 #----------------------------------
 require 'mkmf'
 
@@ -25,6 +24,9 @@ def create_win32ole_makefile
      have_library("kernel32") and
      have_library("advapi32") and
      have_header("windows.h")
+    unless have_type("IMultiLanguage2", "mlang.h")
+      have_type("IMultiLanguage", "mlang.h")
+    end
     create_makefile("win32ole")
     create_docfile(SRCFILES)
   else
@@ -32,8 +34,11 @@ def create_win32ole_makefile
   end
 end
 
+
 case RUBY_PLATFORM
-when /mswin32/
+when /mswin/
   $CFLAGS += ' /W3'
+when /cygwin/, /mingw/
+  $defs << '-DNONAMELESSUNION'
 end
 create_win32ole_makefile
