@@ -2,8 +2,7 @@
 
   iseq.c -
 
-  $Author: ko1 $
-  $Date: 2007-12-25 18:43:23 +0900 (Tue, 25 Dec 2007) $
+  $Author: yugui $
   created at: 2006-07-11(Tue) 09:00:03 +0900
 
   Copyright (C) 2006 Koichi Sasada
@@ -11,19 +10,15 @@
 **********************************************************************/
 
 #include "ruby/ruby.h"
-#include "ruby/node.h"
 
 /* #define MARK_FREE_DEBUG 1 */
 #include "gc.h"
 #include "vm_core.h"
+#include "iseq.h"
 
 #include "insns.inc"
 #include "insns_info.inc"
 
-/* compile.c */
-void iseq_compile(VALUE self, NODE *node);
-int iseq_translate_threaded_code(rb_iseq_t *iseq);
-
 VALUE rb_cISeq;
 
 static void
@@ -49,20 +44,22 @@ iseq_free(void *ptr)
 
     if (ptr) {
 	iseq = ptr;
-	/* It's possible that strings are freed
-         * GC_INFO("%s @ %s\n", RSTRING_PTR(iseq->name),
-         *                      RSTRING_PTR(iseq->filename));
-	 */
-	if (iseq->iseq != iseq->iseq_encoded) {
-	    RUBY_FREE_UNLESS_NULL(iseq->iseq_encoded);
-	}
+	if (!iseq->orig) {
+	    /* It's possible that strings are freed
+	     * GC_INFO("%s @ %s\n", RSTRING_PTR(iseq->name),
+	     *                      RSTRING_PTR(iseq->filename));
+	     */
+	    if (iseq->iseq != iseq->iseq_encoded) {
+		RUBY_FREE_UNLESS_NULL(iseq->iseq_encoded);
+	    }
 
-	RUBY_FREE_UNLESS_NULL(iseq->iseq);
-	RUBY_FREE_UNLESS_NULL(iseq->insn_info_table);
-	RUBY_FREE_UNLESS_NULL(iseq->local_table);
-	RUBY_FREE_UNLESS_NULL(iseq->catch_table);
-	RUBY_FREE_UNLESS_NULL(iseq->arg_opt_table);
-	compile_data_free(iseq->compile_data);
+	    RUBY_FREE_UNLESS_NULL(iseq->iseq);
+	    RUBY_FREE_UNLESS_NULL(iseq->insn_info_table);
+	    RUBY_FREE_UNLESS_NULL(iseq->local_table);
+	    RUBY_FREE_UNLESS_NULL(iseq->catch_table);
+	    RUBY_FREE_UNLESS_NULL(iseq->arg_opt_table);
+	    compile_data_free(iseq->compile_data);
+	}
 	ruby_xfree(ptr);
     }
     RUBY_FREE_LEAVE("iseq");
@@ -82,8 +79,10 @@ iseq_mark(void *ptr)
 	RUBY_MARK_UNLESS_NULL(iseq->filename);
 	RUBY_MARK_UNLESS_NULL((VALUE)iseq->cref_stack);
 	RUBY_MARK_UNLESS_NULL(iseq->klass);
+	RUBY_MARK_UNLESS_NULL(iseq->coverage);
 /* 	RUBY_MARK_UNLESS_NULL((VALUE)iseq->node); */
 /*	RUBY_MARK_UNLESS_NULL(iseq->cached_special_block); */
+	RUBY_MARK_UNLESS_NULL(iseq->orig);
 
 	if (iseq->compile_data != 0) {
 	    RUBY_MARK_UNLESS_NULL(iseq->compile_data->mark_ary);
@@ -192,6 +191,16 @@ prepare_iseq_build(rb_iseq_t *iseq,
 
     set_relation(iseq, parent);
 
+    iseq->coverage = Qfalse;
+    if (!GET_THREAD()->parse_in_eval) {
+	extern VALUE rb_get_coverages(void);
+	VALUE coverages = rb_get_coverages();
+	if (RTEST(coverages)) {
+	    iseq->coverage = rb_hash_lookup(coverages, filename);
+	    if (NIL_P(iseq->coverage)) iseq->coverage = Qfalse;
+	}
+    }
+
     return Qtrue;
 }
 
@@ -220,7 +229,7 @@ static rb_compile_option_t COMPILE_OPTION_DEFAULT = {
     OPT_STACK_CACHING, /* int stack_caching; */
     OPT_TRACE_INSTRUCTION, /* int trace_instruction */
 };
-static const rb_compile_option_t COMPILE_OPTION_FALSE;
+static const rb_compile_option_t COMPILE_OPTION_FALSE = {0};
 
 static void
 make_compile_option(rb_compile_option_t *option, VALUE opt)
@@ -241,6 +250,10 @@ make_compile_option(rb_compile_option_t *option, VALUE opt)
   { VALUE flag = rb_hash_aref(h, ID2SYM(rb_intern(#mem))); \
       if (flag == Qtrue)  { o->mem = 1; } \
       else if (flag == Qfalse)  { o->mem = 0; } \
+  }
+#define SET_COMPILE_OPTION_NUM(o, h, mem) \
+  { VALUE num = rb_hash_aref(opt, ID2SYM(rb_intern(#mem))); \
+      if (!NIL_P(num)) o->mem = NUM2INT(num); \
   }
 	SET_COMPILE_OPTION(option, opt, inline_const_cache);
 	SET_COMPILE_OPTION(option, opt, peephole_optimization);
@@ -250,7 +263,9 @@ make_compile_option(rb_compile_option_t *option, VALUE opt)
 	SET_COMPILE_OPTION(option, opt, instructions_unification);
 	SET_COMPILE_OPTION(option, opt, stack_caching);
 	SET_COMPILE_OPTION(option, opt, trace_instruction);
+	SET_COMPILE_OPTION_NUM(option, opt, debug_level);
 #undef SET_COMPILE_OPTION
+#undef SET_COMPILE_OPTION_NUM
     }
     else {
 	rb_raise(rb_eTypeError, "Compile option must be Hash/true/false/nil");
@@ -263,6 +278,8 @@ make_compile_option_value(rb_compile_option_t *option)
     VALUE opt = rb_hash_new();
 #define SET_COMPILE_OPTION(o, h, mem) \
   rb_hash_aset(h, ID2SYM(rb_intern(#mem)), o->mem ? Qtrue : Qfalse)
+#define SET_COMPILE_OPTION_NUM(o, h, mem) \
+  rb_hash_aset(h, ID2SYM(rb_intern(#mem)), INT2NUM(o->mem))
     {
 	SET_COMPILE_OPTION(option, opt, inline_const_cache);
 	SET_COMPILE_OPTION(option, opt, peephole_optimization);
@@ -271,8 +288,10 @@ make_compile_option_value(rb_compile_option_t *option)
 	SET_COMPILE_OPTION(option, opt, operands_unification);
 	SET_COMPILE_OPTION(option, opt, instructions_unification);
 	SET_COMPILE_OPTION(option, opt, stack_caching);
+	SET_COMPILE_OPTION_NUM(option, opt, debug_level);
     }
 #undef SET_COMPILE_OPTION
+#undef SET_COMPILE_OPTION_NUM
     return opt;
 }
 
@@ -284,6 +303,22 @@ rb_iseq_new(NODE *node, VALUE name, VALUE filename,
 				&COMPILE_OPTION_DEFAULT);
 }
 
+VALUE
+rb_iseq_new_top(NODE *node, VALUE name, VALUE filename, VALUE parent)
+{
+    return rb_iseq_new_with_opt(node, name, filename, parent, ISEQ_TYPE_TOP,
+				&COMPILE_OPTION_DEFAULT);
+}
+
+VALUE
+rb_iseq_new_main(NODE *node, VALUE filename)
+{
+    rb_thread_t *th = GET_THREAD();
+    VALUE parent = th->base_block->iseq->self;
+    return rb_iseq_new_with_opt(node, rb_str_new2("<main>"), filename,
+				parent, ISEQ_TYPE_MAIN, &COMPILE_OPTION_DEFAULT);
+}
+
 static VALUE
 rb_iseq_new_with_bopt_and_opt(NODE *node, VALUE name, VALUE filename,
 				VALUE parent, VALUE type, VALUE bopt,
@@ -296,7 +331,7 @@ rb_iseq_new_with_bopt_and_opt(NODE *node, VALUE name, VALUE filename,
     iseq->self = self;
 
     prepare_iseq_build(iseq, name, filename, parent, type, bopt, option);
-    iseq_compile(self, node);
+    rb_iseq_compile_node(self, node);
     cleanup_iseq_build(iseq);
     return self;
 }
@@ -318,54 +353,51 @@ rb_iseq_new_with_bopt(NODE *node, VALUE name, VALUE filename,
 					   bopt, &COMPILE_OPTION_DEFAULT);
 }
 
-VALUE iseq_build_from_ary(rb_iseq_t *iseq, VALUE locals, VALUE args,
-			  VALUE exception, VALUE body);
-
 #define CHECK_ARRAY(v)   rb_convert_type(v, T_ARRAY, "Array", "to_ary")
 #define CHECK_STRING(v)  rb_convert_type(v, T_STRING, "String", "to_str")
 #define CHECK_SYMBOL(v)  rb_convert_type(v, T_SYMBOL, "Symbol", "to_sym")
-#define CHECK_INTEGER(v) (NUM2LONG(v), v)
-VALUE
+static inline VALUE CHECK_INTEGER(VALUE v) {NUM2LONG(v); return v;}
+static VALUE
 iseq_load(VALUE self, VALUE data, VALUE parent, VALUE opt)
 {
-    VALUE iseqval = iseq_alloc(rb_cISeq);
+    VALUE iseqval = iseq_alloc(self);
 
     VALUE magic, version1, version2, format_type, misc;
-    VALUE name, filename, line;
+    VALUE name, filename;
     VALUE type, body, locals, args, exception;
 
     VALUE iseq_type;
     struct st_table *type_map = 0;
     rb_iseq_t *iseq;
     rb_compile_option_t option;
+    int i = 0;
 
     /* [magic, major_version, minor_version, format_type, misc,
-     *  name, filename, line,
+     *  name, filename,
      *  type, locals, args, exception_table, body]
      */
 
     data        = CHECK_ARRAY(data);
 
-    magic       = CHECK_STRING(rb_ary_entry(data, 0));
-    version1    = CHECK_INTEGER(rb_ary_entry(data, 1));
-    version2    = CHECK_INTEGER(rb_ary_entry(data, 2));
-    format_type = CHECK_INTEGER(rb_ary_entry(data, 3));
-    misc        = rb_ary_entry(data, 4); /* TODO */
+    magic       = CHECK_STRING(rb_ary_entry(data, i++));
+    version1    = CHECK_INTEGER(rb_ary_entry(data, i++));
+    version2    = CHECK_INTEGER(rb_ary_entry(data, i++));
+    format_type = CHECK_INTEGER(rb_ary_entry(data, i++));
+    misc        = rb_ary_entry(data, i++); /* TODO */
 
-    name        = CHECK_STRING(rb_ary_entry(data, 5));
-    filename    = CHECK_STRING(rb_ary_entry(data, 6));
-    line        = CHECK_ARRAY(rb_ary_entry(data, 7));
+    name        = CHECK_STRING(rb_ary_entry(data, i++));
+    filename    = CHECK_STRING(rb_ary_entry(data, i++));
 
-    type        = CHECK_SYMBOL(rb_ary_entry(data, 8));
-    locals      = CHECK_ARRAY(rb_ary_entry(data, 9));
+    type        = CHECK_SYMBOL(rb_ary_entry(data, i++));
+    locals      = CHECK_ARRAY(rb_ary_entry(data, i++));
 
-    args        = rb_ary_entry(data, 10);
+    args        = rb_ary_entry(data, i++);
     if (FIXNUM_P(args) || (args = CHECK_ARRAY(args))) {
 	/* */
     }
 
-    exception   = CHECK_ARRAY(rb_ary_entry(data, 11));
-    body        = CHECK_ARRAY(rb_ary_entry(data, 12));
+    exception   = CHECK_ARRAY(rb_ary_entry(data, i++));
+    body        = CHECK_ARRAY(rb_ary_entry(data, i++));
 
     GetISeqPtr(iseqval, iseq);
     iseq->self = iseqval;
@@ -379,6 +411,8 @@ iseq_load(VALUE self, VALUE data, VALUE parent, VALUE opt)
 	st_insert(type_map, ID2SYM(rb_intern("rescue")), ISEQ_TYPE_RESCUE);
 	st_insert(type_map, ID2SYM(rb_intern("ensure")), ISEQ_TYPE_ENSURE);
 	st_insert(type_map, ID2SYM(rb_intern("eval")), ISEQ_TYPE_EVAL);
+	st_insert(type_map, ID2SYM(rb_intern("main")), ISEQ_TYPE_MAIN);
+	st_insert(type_map, ID2SYM(rb_intern("defined_guard")), ISEQ_TYPE_DEFINED_GUARD);
     }
 
     if (st_lookup(type_map, type, &iseq_type) == 0) {
@@ -397,7 +431,7 @@ iseq_load(VALUE self, VALUE data, VALUE parent, VALUE opt)
     prepare_iseq_build(iseq, name, filename,
 		       parent, iseq_type, 0, &option);
 
-    iseq_build_from_ary(iseq, locals, args, exception, body);
+    rb_iseq_build_from_ary(iseq, locals, args, exception, body);
 
     cleanup_iseq_build(iseq);
     return iseqval;
@@ -412,6 +446,12 @@ iseq_s_load(int argc, VALUE *argv, VALUE self)
     return iseq_load(self, data, 0, opt);
 }
 
+VALUE
+ruby_iseq_load(VALUE data, VALUE parent, VALUE opt)
+{
+    return iseq_load(rb_cISeq, data, parent, opt);
+}
+
 static NODE *
 compile_string(VALUE str, VALUE file, VALUE line)
 {
@@ -429,11 +469,11 @@ VALUE
 rb_iseq_compile_with_option(VALUE src, VALUE file, VALUE line, VALUE opt)
 {
     rb_compile_option_t option;
-    NODE *node = compile_string(src, file, line);
+    NODE *node = compile_string(StringValue(src), file, line);
     rb_thread_t *th = GET_THREAD();
     make_compile_option(&option, opt);
 
-    if (th->base_block) {
+    if (th->base_block && th->base_block->iseq) {
 	return rb_iseq_new_with_opt(node, th->base_block->iseq->name,
 				    file, th->base_block->iseq->self,
 				    ISEQ_TYPE_EVAL, &option);
@@ -476,9 +516,10 @@ iseq_s_compile_file(int argc, VALUE *argv, VALUE self)
 
     rb_secure(1);
     rb_scan_args(argc, argv, "11", &file, &opt);
+    FilePathValue(file);
     fname = StringValueCStr(file);
 
-    f = rb_file_open(fname, "r");
+    f = rb_file_open_str(file, "r");
 
     parser = rb_parser_new();
     node = rb_parser_compile_file(parser, fname, f, NUM2INT(line));
@@ -524,15 +565,18 @@ iseq_eval(VALUE self)
 static VALUE
 iseq_inspect(VALUE self)
 {
-    char buff[0x100];
-    rb_iseq_t *iseq = iseq_check(self);
+    rb_iseq_t *iseq;
+    GetISeqPtr(self, iseq);
+    if (!iseq->name) {
+        return rb_sprintf("#<%s: uninitialized>", rb_obj_classname(self));
+    }
 
-    snprintf(buff, sizeof(buff), "<ISeq:%s@%s>",
-	     RSTRING_PTR(iseq->name), RSTRING_PTR(iseq->filename));
-
-    return rb_str_new2(buff);
+    return rb_sprintf("<%s:%s@%s>",
+                      rb_obj_classname(self),
+		      RSTRING_PTR(iseq->name), RSTRING_PTR(iseq->filename));
 }
 
+static
 VALUE iseq_data_to_ary(rb_iseq_t *iseq);
 
 static VALUE
@@ -543,6 +587,12 @@ iseq_to_a(VALUE self)
     return iseq_data_to_ary(iseq);
 }
 
+int
+rb_iseq_first_lineno(rb_iseq_t *iseq)
+{
+    return iseq->insn_info_table[0].line_no;
+}
+
 /* TODO: search algorithm is brute force.
          this should be binary search or so. */
 
@@ -598,20 +648,17 @@ insn_operand_intern(rb_iseq_t *iseq,
 		    int insn, int op_no, VALUE op,
 		    int len, int pos, VALUE *pnop, VALUE child)
 {
-    char *types = insn_op_types(insn);
+    const char *types = insn_op_types(insn);
     char type = types[op_no];
     VALUE ret;
-    char buff[0x100];
 
     switch (type) {
       case TS_OFFSET:		/* LONG */
-	snprintf(buff, sizeof(buff), "%ld", pos + len + op);
-	ret = rb_str_new2(buff);
+	ret = rb_sprintf("%ld", pos + len + op);
 	break;
 
       case TS_NUM:		/* ULONG */
-	snprintf(buff, sizeof(buff), "%lu", op);
-	ret = rb_str_new2(buff);
+	ret = rb_sprintf("%lu", op);
 	break;
 
       case TS_LINDEX:
@@ -650,6 +697,7 @@ insn_operand_intern(rb_iseq_t *iseq,
       }
       case TS_ID:		/* ID (symbol) */
 	op = ID2SYM(op);
+
       case TS_VALUE:		/* VALUE */
 	ret = rb_inspect(op);
 	if (CLASS_OF(op) == rb_cISeq) {
@@ -691,7 +739,7 @@ insn_operand_intern(rb_iseq_t *iseq,
 	break;
 
       default:
-	rb_bug("ruby_iseq_disasm: unknown operand type: %c", type);
+	rb_bug("rb_iseq_disasm: unknown operand type: %c", type);
     }
     return ret;
 }
@@ -701,31 +749,27 @@ insn_operand_intern(rb_iseq_t *iseq,
  * Iseq -> Iseq inspect object
  */
 VALUE
-ruby_iseq_disasm_insn(VALUE ret, VALUE *iseq, int pos,
-		      rb_iseq_t *iseqdat, VALUE child)
+rb_iseq_disasm_insn(VALUE ret, VALUE *iseq, int pos,
+		    rb_iseq_t *iseqdat, VALUE child)
 {
     int insn = iseq[pos];
     int len = insn_len(insn);
-    int i, j;
-    char *types = insn_op_types(insn);
+    int j;
+    const char *types = insn_op_types(insn);
     VALUE str = rb_str_new(0, 0);
-    char buff[0x100];
-    char insn_name_buff[0x100];
+    const char *insn_name_buff;
 
-    strcpy(insn_name_buff, insn_name(insn));
-    if (0) {
-	for (i = 0; insn_name_buff[i]; i++) {
-	    if (insn_name_buff[i] == '_') {
-		insn_name_buff[i] = 0;
-	    }
-	}
+    insn_name_buff = insn_name(insn);
+    if (1) {
+	rb_str_catf(str, "%04d %-16s ", pos, insn_name_buff);
+    }
+    else {
+	rb_str_catf(str, "%04d %-16.*s ", pos,
+		    (int)strcspn(insn_name_buff, "_"), insn_name_buff);
     }
 
-    snprintf(buff, sizeof(buff), "%04d %-16s ", pos, insn_name_buff);
-    rb_str_cat2(str, buff);
-
     for (j = 0; types[j]; j++) {
-	char *types = insn_op_types(insn);
+	const char *types = insn_op_types(insn);
 	VALUE opstr = insn_operand_intern(iseqdat, insn, j, iseq[pos + j + 1],
 					  len, pos, &iseq[pos + j + 2],
 					  child);
@@ -740,17 +784,18 @@ ruby_iseq_disasm_insn(VALUE ret, VALUE *iseq, int pos,
 	int line_no = find_line_no(iseqdat, pos);
 	int prev = find_prev_line_no(iseqdat, pos);
 	if (line_no && line_no != prev) {
-	    snprintf(buff, sizeof(buff), "%-70s(%4d)", RSTRING_PTR(str),
-		     line_no);
-	    str = rb_str_new2(buff);
+	    long slen = RSTRING_LEN(str);
+	    slen = (slen > 70) ? 0 : (70 - slen);
+	    str = rb_str_catf(str, "%*s(%4d)", (int)slen, "", line_no);
 	}
     }
     else {
 	/* for debug */
 	struct iseq_insn_info_entry *entry = get_insn_info(iseqdat, pos);
-	snprintf(buff, sizeof(buff), "%-60s(line: %d, sp: %d)",
-		 RSTRING_PTR(str), entry->line_no, entry->sp);
-	str = rb_str_new2(buff);
+	long slen = RSTRING_LEN(str);
+	slen = (slen > 60) ? 0 : (60 - slen);
+	str = rb_str_catf(str, "%*s(line: %d, sp: %d)",
+			  (int)slen, "", entry->line_no, entry->sp);
     }
 
     if (ret) {
@@ -763,7 +808,7 @@ ruby_iseq_disasm_insn(VALUE ret, VALUE *iseq, int pos,
     return len;
 }
 
-static char *
+static const char *
 catch_type(int type)
 {
     switch (type) {
@@ -786,7 +831,7 @@ catch_type(int type)
 }
 
 VALUE
-ruby_iseq_disasm(VALUE self)
+rb_iseq_disasm(VALUE self)
 {
     rb_iseq_t *iseqdat = iseq_check(self);
     VALUE *iseq;
@@ -795,7 +840,6 @@ ruby_iseq_disasm(VALUE self)
     unsigned long size;
     int i;
     ID *tbl;
-    char buff[0x200];
     enum {header_minlen = 72};
 
     rb_secure(1);
@@ -818,13 +862,12 @@ ruby_iseq_disasm(VALUE self)
     }
     for (i = 0; i < iseqdat->catch_table_size; i++) {
 	struct iseq_catch_table_entry *entry = &iseqdat->catch_table[i];
-	sprintf(buff,
-		"| catch type: %-6s st: %04d ed: %04d sp: %04d cont: %04d\n",
-		catch_type((int)entry->type), (int)entry->start,
-		(int)entry->end, (int)entry->sp, (int)entry->cont);
-	rb_str_cat2(str, buff);
+	rb_str_catf(str,
+		    "| catch type: %-6s st: %04d ed: %04d sp: %04d cont: %04d\n",
+		    catch_type((int)entry->type), (int)entry->start,
+		    (int)entry->end, (int)entry->sp, (int)entry->cont);
 	if (entry->iseq) {
-	    rb_str_concat(str, ruby_iseq_disasm(entry->iseq));
+	    rb_str_concat(str, rb_iseq_disasm(entry->iseq));
 	}
     }
     if (iseqdat->catch_table_size != 0) {
@@ -836,14 +879,13 @@ ruby_iseq_disasm(VALUE self)
     tbl = iseqdat->local_table;
 
     if (tbl) {
-	snprintf(buff, sizeof(buff),
-		 "local table (size: %d, argc: %d "
-		 "[opts: %d, rest: %d, post: %d, block: %d] s%d)\n",
-		 iseqdat->local_size, iseqdat->argc,
-		 iseqdat->arg_opts, iseqdat->arg_rest,
-		 iseqdat->arg_post_len, iseqdat->arg_block,
-		 iseqdat->arg_simple);
-	rb_str_cat2(str, buff);
+	rb_str_catf(str,
+		    "local table (size: %d, argc: %d "
+		    "[opts: %d, rest: %d, post: %d, block: %d] s%d)\n",
+		    iseqdat->local_size, iseqdat->argc,
+		    iseqdat->arg_opts, iseqdat->arg_rest,
+		    iseqdat->arg_post_len, iseqdat->arg_block,
+		    iseqdat->arg_simple);
 
 	for (i = 0; i < iseqdat->local_table_size; i++) {
 	    const char *name = rb_id2name(tbl[i]);
@@ -871,22 +913,19 @@ ruby_iseq_disasm(VALUE self)
 	    snprintf(info, sizeof(info), "%s%s%s%s", name ? name : "?",
 		     *argi ? "<" : "", argi, *argi ? ">" : "");
 
-	    snprintf(buff, sizeof(buff), "[%2d] %-11s",
-		     iseqdat->local_size - i, info);
-
-	    rb_str_cat2(str, buff);
+	    rb_str_catf(str, "[%2d] %-11s", iseqdat->local_size - i, info);
 	}
 	rb_str_cat2(str, "\n");
     }
 
     /* show each line */
     for (i = 0; i < size;) {
-	i += ruby_iseq_disasm_insn(str, iseq, i, iseqdat, child);
+	i += rb_iseq_disasm_insn(str, iseq, i, iseqdat, child);
     }
 
     for (i = 0; i < RARRAY_LEN(child); i++) {
 	VALUE isv = rb_ary_entry(child, i);
-	rb_str_concat(str, ruby_iseq_disasm(isv));
+	rb_str_concat(str, rb_iseq_disasm(isv));
     }
 
     return str;
@@ -903,8 +942,8 @@ iseq_s_disasm(VALUE klass, VALUE body)
 
     if ((node = rb_method_body(body)) != 0) {
 	if (nd_type(node) == RUBY_VM_METHOD_NODE) {
-	    VALUE iseqval = (VALUE)node->nd_body;
-	    ret = ruby_iseq_disasm(iseqval);
+ 	    VALUE iseqval = (VALUE)node->nd_body;
+	    ret = rb_iseq_disasm(iseqval);
 	}
     }
 
@@ -932,9 +971,9 @@ static VALUE
 register_label(struct st_table *table, int idx)
 {
     VALUE sym;
-    char buff[0x20];
+    char buff[8 + (sizeof(idx) * CHAR_BIT * 32 / 100)];
 
-    snprintf(buff, 0x20, "label_%u", idx);
+    snprintf(buff, sizeof(buff), "label_%u", idx);
     sym = ID2SYM(rb_intern(buff));
     st_insert(table, idx, sym);
     return sym;
@@ -945,12 +984,12 @@ exception_type2symbol(VALUE type)
 {
     ID id;
     switch(type) {
-      case CATCH_TYPE_RESCUE: id = rb_intern("rescue"); break;
-      case CATCH_TYPE_ENSURE: id = rb_intern("ensure"); break;
-      case CATCH_TYPE_RETRY:  id = rb_intern("retry");  break;
-      case CATCH_TYPE_BREAK:  id = rb_intern("break");  break;
-      case CATCH_TYPE_REDO:   id = rb_intern("redo");   break;
-      case CATCH_TYPE_NEXT:   id = rb_intern("next");   break;
+      case CATCH_TYPE_RESCUE: CONST_ID(id, "rescue"); break;
+      case CATCH_TYPE_ENSURE: CONST_ID(id, "ensure"); break;
+      case CATCH_TYPE_RETRY:  CONST_ID(id, "retry");  break;
+      case CATCH_TYPE_BREAK:  CONST_ID(id, "break");  break;
+      case CATCH_TYPE_REDO:   CONST_ID(id, "redo");   break;
+      case CATCH_TYPE_NEXT:   CONST_ID(id, "next");   break;
       default:
 	rb_bug("...");
     }
@@ -965,7 +1004,7 @@ cdhash_each(VALUE key, VALUE value, VALUE ary)
     return ST_CONTINUE;
 }
 
-VALUE
+static VALUE
 iseq_data_to_ary(rb_iseq_t *iseq)
 {
     int i, pos, line = 0;
@@ -990,6 +1029,8 @@ iseq_data_to_ary(rb_iseq_t *iseq)
     DECL_SYMBOL(rescue);
     DECL_SYMBOL(ensure);
     DECL_SYMBOL(eval);
+    DECL_SYMBOL(main);
+    DECL_SYMBOL(defined_guard);
 
     if (sym_top == 0) {
 	int i;
@@ -1003,6 +1044,8 @@ iseq_data_to_ary(rb_iseq_t *iseq)
 	INIT_SYMBOL(rescue);
 	INIT_SYMBOL(ensure);
 	INIT_SYMBOL(eval);
+	INIT_SYMBOL(main);
+	INIT_SYMBOL(defined_guard);
     }
 
     /* type */
@@ -1014,6 +1057,8 @@ iseq_data_to_ary(rb_iseq_t *iseq)
       case ISEQ_TYPE_RESCUE: type = sym_rescue; break;
       case ISEQ_TYPE_ENSURE: type = sym_ensure; break;
       case ISEQ_TYPE_EVAL:   type = sym_eval;   break;
+      case ISEQ_TYPE_MAIN:   type = sym_main;   break;
+      case ISEQ_TYPE_DEFINED_GUARD: type = sym_defined_guard; break;
       default: rb_bug("unsupported iseq type");
     };
 
@@ -1203,18 +1248,28 @@ iseq_data_to_ary(rb_iseq_t *iseq)
     return val;
 }
 
-struct st_table *
-insn_make_insn_table(void)
+VALUE
+rb_iseq_clone(VALUE iseqval, VALUE newcbase)
 {
-    struct st_table *table;
-    int i;
-    table = st_init_numtable();
+    VALUE newiseq = iseq_alloc(rb_cISeq);
+    rb_iseq_t *iseq0, *iseq1;
 
-    for (i=0; i<VM_INSTRUCTION_SIZE; i++) {
-	st_insert(table, ID2SYM(rb_intern(insn_name(i))), i);
+    GetISeqPtr(iseqval, iseq0);
+    GetISeqPtr(newiseq, iseq1);
+
+    *iseq1 = *iseq0;
+    iseq1->self = newiseq;
+    if (!iseq1->orig) {
+	iseq1->orig = iseqval;
+    }
+    if (newcbase) {
+	iseq1->cref_stack = NEW_BLOCK(newcbase);
+	if (iseq0->cref_stack->nd_next) {
+	    iseq1->cref_stack->nd_next = iseq0->cref_stack->nd_next;
+	}
     }
 
-    return table;
+    return newiseq;
 }
 
 /* ruby2cext */
@@ -1249,7 +1304,7 @@ rb_iseq_build_for_ruby2cext(
 	iseq->iseq[i+1] = (VALUE)func;
     }
 
-    iseq_translate_threaded_code(iseq);
+    rb_iseq_translate_threaded_code(iseq);
 
 #define ALLOC_AND_COPY(dst, src, type, size) do { \
   if (size) { \
@@ -1276,16 +1331,17 @@ void
 Init_ISeq(void)
 {
     /* declare ::VM::InstructionSequence */
-    rb_cISeq = rb_define_class_under(rb_cVM, "InstructionSequence", rb_cObject);
+    rb_cISeq = rb_define_class_under(rb_cRubyVM, "InstructionSequence", rb_cObject);
     rb_define_alloc_func(rb_cISeq, iseq_alloc);
     rb_define_method(rb_cISeq, "inspect", iseq_inspect, 0);
-    rb_define_method(rb_cISeq, "disasm", ruby_iseq_disasm, 0);
-    rb_define_method(rb_cISeq, "disassemble", ruby_iseq_disasm, 0);
+    rb_define_method(rb_cISeq, "disasm", rb_iseq_disasm, 0);
+    rb_define_method(rb_cISeq, "disassemble", rb_iseq_disasm, 0);
     rb_define_method(rb_cISeq, "to_a", iseq_to_a, 0);
     rb_define_method(rb_cISeq, "eval", iseq_eval, 0);
 
     /* disable this feature because there is no verifier. */
     /* rb_define_singleton_method(rb_cISeq, "load", iseq_s_load, -1); */
+    (void)iseq_s_load;
 
     rb_define_singleton_method(rb_cISeq, "compile", iseq_s_compile, -1);
     rb_define_singleton_method(rb_cISeq, "new", iseq_s_compile, -1);
