@@ -1,8 +1,7 @@
 /* public domain rewrite of finite(3) */
 
 int
-finite(n)
-    double n;
+finite(double n)
 {
     return !isnan(n) && !isinf(n);
 }
