@@ -2,7 +2,7 @@
 
   range.c -
 
-  $Author: nobu $
+  $Author: yugui $
   created at: Thu Aug 19 17:46:47 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -267,6 +267,8 @@ step_i(VALUE i, void *arg)
     return Qnil;
 }
 
+extern int ruby_float_step(VALUE from, VALUE to, VALUE step, int excl);
+
 /*
  *  call-seq:
  *     rng.step(n=1) {| obj | block }    => rng
@@ -334,6 +336,9 @@ range_step(int argc, VALUE *argv, VALUE range)
 	}
 
     }
+    else if (ruby_float_step(b, e, step, EXCL(range))) {
+	/* done */
+    }
     else if (rb_obj_is_kind_of(b, rb_cNumeric) ||
 	     !NIL_P(rb_check_to_integer(b, "to_int")) ||
 	     !NIL_P(rb_check_to_integer(e, "to_int"))) {
@@ -636,7 +641,7 @@ rb_range_beg_len(VALUE range, long *begp, long *lenp, long len, int err)
   out_of_range:
     if (err) {
 	rb_raise(rb_eRangeError, "%ld..%s%ld out of range",
-		 b, excl ? "." : "", e);
+		 NUM2LONG(b), excl ? "." : "", NUM2LONG(e));
     }
     return Qnil;
 }
@@ -663,6 +668,24 @@ range_to_s(VALUE range)
     return str;
 }
 
+static VALUE
+inspect_range(VALUE range, VALUE dummy, int recur)
+{
+    VALUE str, str2;
+
+    if (recur) {
+	return rb_str_new2(EXCL(range) ? "(... ... ...)" : "(... .. ...)");
+    }
+    str = rb_inspect(RANGE_BEG(range));
+    str2 = rb_inspect(RANGE_END(range));
+    str = rb_str_dup(str);
+    rb_str_cat(str, "...", EXCL(range) ? 3 : 2);
+    rb_str_append(str, str2);
+    OBJ_INFECT(str, str2);
+
+    return str;
+}
+
 /*
  * call-seq:
  *   rng.inspect  => string
@@ -676,16 +699,7 @@ range_to_s(VALUE range)
 static VALUE
 range_inspect(VALUE range)
 {
-    VALUE str, str2;
-
-    str = rb_inspect(RANGE_BEG(range));
-    str2 = rb_inspect(RANGE_END(range));
-    str = rb_str_dup(str);
-    rb_str_cat(str, "...", EXCL(range) ? 3 : 2);
-    rb_str_append(str, str2);
-    OBJ_INFECT(str, str2);
-
-    return str;
+    return rb_exec_recursive(inspect_range, range, 0);
 }
 
 /*
