@@ -57,7 +57,7 @@
 #include <sys/types.h>
 #include <errno.h>
 #endif
-#if defined(TM_IN_SYS_TIME) || !defined(GAWK) && !defined(_WIN32_WCE)
+#if defined(TM_IN_SYS_TIME) || !defined(GAWK)
 #include <sys/types.h>
 #if HAVE_SYS_TIME_H
 #include <sys/time.h>
@@ -126,7 +126,7 @@ extern char *strchr();
 #ifndef DLL_IMPORT
 #define DLL_IMPORT
 #endif
-#if !defined(OS2) && !defined(MSDOS) && defined(HAVE_TZNAME)
+#if !defined(OS2) && defined(HAVE_TZNAME)
 extern DLL_IMPORT char *tzname[2];
 #ifdef HAVE_DAYLIGHT
 extern DLL_IMPORT int daylight;
@@ -195,12 +195,11 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 #endif /* POSIX_SEMANTICS */
 #ifndef HAVE_TM_ZONE
 #ifndef HAVE_TM_NAME
-#if !defined HAVE_VAR_TIMEZONE || defined HAVE_TIMEZONE
+#if ((defined(MAILHEADER_EXT) && !HAVE_VAR_TIMEZONE && HAVE_GETTIMEOFDAY) || \
+     (!HAVE_TZNAME && HAVE_TIMEZONE))
 	struct timeval tv;
-#endif
-#ifdef HAVE_TIMEZONE
 	struct timezone zone;
-#endif /* HAVE_TIMEZONE */
+#endif
 #endif /* HAVE_TM_NAME */
 #endif /* HAVE_TM_ZONE */
 	int precision, flags;
@@ -271,22 +270,37 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 				goto unknown; \
 		} while (0)
 #define NEEDS(n) do if (s + (n) >= endp - 1) goto err; while (0)
+#define FILL_PADDING(i) do { \
+	if (!(flags & BIT_OF(LEFT)) && precision > i) { \
+		NEEDS(precision); \
+		memset(s, padding ? padding : ' ', precision - i); \
+		s += precision - i; \
+	} \
+	else { \
+		NEEDS(i); \
+	} \
+} while (0);
 #define FMT(def_pad, def_prec, fmt, val) \
 		do { \
 			int l; \
 			if (precision <= 0) precision = (def_prec); \
-			if (flags & BIT_OF(LEFT)) precision = 0; \
+			if (flags & BIT_OF(LEFT)) precision = 1; \
 			l = snprintf(s, endp - s, \
-				     ((!padding || padding == (def_pad)) ? "%.*"fmt : "%*"fmt), \
+				     ((padding == '0' || (!padding && def_pad == '0')) ? "%0*"fmt : "%*"fmt), \
 				     precision, val); \
 			if (l < 0) goto err; \
 			s += l; \
 		} while (0)
-#define STRFTIME(fmt, tm) \
+#define STRFTIME(fmt) \
 		do { \
-			i = strftime(s, endp - s, fmt, tm); \
+			i = rb_strftime(s, endp - s, fmt, timeptr, ts, gmt); \
 			if (!i) return 0; \
-			s += i; \
+			if (precision > i) {\
+				memmove(s + precision - i, s, i);\
+				memset(s, padding ? padding : ' ', precision - i); \
+				s += precision;	\
+	                }\
+			else s += i; \
 		} while (0)
 
 		if (*format != '%') {
@@ -301,14 +315,19 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 	again:
 		switch (*++format) {
 		case '\0':
-			*s++ = '%';
-			goto out;
+			format--;
+			goto unknown;
 
 		case '%':
+			FILL_PADDING(1);
 			*s++ = '%';
 			continue;
 
 		case 'a':	/* abbreviated weekday name */
+			if (flags & BIT_OF(CHCASE)) {
+				flags &= ~(BIT_OF(LOWER)|BIT_OF(CHCASE));
+				flags |= BIT_OF(UPPER);
+			}
 			if (timeptr->tm_wday < 0 || timeptr->tm_wday > 6)
 				i = 1, tp = "?";
 			else
@@ -316,6 +335,10 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			break;
 
 		case 'A':	/* full weekday name */
+			if (flags & BIT_OF(CHCASE)) {
+				flags &= ~(BIT_OF(LOWER)|BIT_OF(CHCASE));
+				flags |= BIT_OF(UPPER);
+			}
 			if (timeptr->tm_wday < 0 || timeptr->tm_wday > 6)
 				i = 1, tp = "?";
 			else
@@ -326,6 +349,10 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 		case 'h':	/* abbreviated month name */
 #endif
 		case 'b':	/* abbreviated month name */
+			if (flags & BIT_OF(CHCASE)) {
+				flags &= ~(BIT_OF(LOWER)|BIT_OF(CHCASE));
+				flags |= BIT_OF(UPPER);
+			}
 			if (timeptr->tm_mon < 0 || timeptr->tm_mon > 11)
 				i = 1, tp = "?";
 			else
@@ -333,6 +360,10 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			break;
 
 		case 'B':	/* full month name */
+			if (flags & BIT_OF(CHCASE)) {
+				flags &= ~(BIT_OF(LOWER)|BIT_OF(CHCASE));
+				flags |= BIT_OF(UPPER);
+			}
 			if (timeptr->tm_mon < 0 || timeptr->tm_mon > 11)
 				i = 1, tp = "?";
 			else
@@ -340,7 +371,7 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			break;
 
 		case 'c':	/* appropriate date and time representation */
-			STRFTIME("%a %b %e %H:%M:%S %Y", timeptr);
+			STRFTIME("%a %b %e %H:%M:%S %Y");
 			continue;
 
 		case 'd':	/* day of the month, 01 - 31 */
@@ -376,8 +407,10 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			FMT('0', 2, "d", i);
 			continue;
 
-		case 'p':	/* am or pm based on 12-hour clock */
-			if (flags & BIT_OF(CHCASE)) {
+		case 'p':	/* AM or PM based on 12-hour clock */
+		case 'P':	/* am or pm based on 12-hour clock */
+			if ((*format == 'p' && (flags & BIT_OF(CHCASE))) ||
+			    (*format == 'P' && !(flags & (BIT_OF(CHCASE)|BIT_OF(UPPER))))) {
 				flags &= ~(BIT_OF(UPPER)|BIT_OF(CHCASE));
 				flags |= BIT_OF(LOWER);
 			}
@@ -389,6 +422,10 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			i = 2;
 			break;
 
+		case 's':
+			FMT('0', 1, "d", (int) ts->tv_sec);
+			continue;
+
 		case 'S':	/* second, 00 - 60 */
 			i = range(0, timeptr->tm_sec, 60);
 			FMT('0', 2, "d", i);
@@ -400,7 +437,7 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 
 		case 'w':	/* weekday, Sunday == 0, 0 - 6 */
 			i = range(0, timeptr->tm_wday, 6);
-			FMT('0', 0, "d", i);
+			FMT('0', 1, "d", i);
 			continue;
 
 		case 'W':	/* week of year, Monday is first day of week */
@@ -408,11 +445,11 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			continue;
 
 		case 'x':	/* appropriate date representation */
-			STRFTIME("%m/%d/%y", timeptr);
+			STRFTIME("%m/%d/%y");
 			continue;
 
 		case 'X':	/* appropriate time representation */
-			STRFTIME("%H:%M:%S", timeptr);
+			STRFTIME("%H:%M:%S");
 			continue;
 
 		case 'y':	/* year without a century, 00 - 99 */
@@ -421,7 +458,7 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			continue;
 
 		case 'Y':	/* year with century */
-			FMT('0', 0, "ld", 1900L + timeptr->tm_year);
+			FMT('0', 1, "ld", 1900L + timeptr->tm_year);
 			continue;
 
 #ifdef MAILHEADER_EXT
@@ -467,12 +504,21 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 #else
 				off = -timezone / 60;
 #endif
-#else /* !HAVE_TIMEZONE */
+#else /* !HAVE_VAR_TIMEZONE */
 #ifdef HAVE_GETTIMEOFDAY
 				gettimeofday(&tv, &zone);
 				off = -zone.tz_minuteswest;
+#else
+				/* no timezone info, then calc by myself */
+				{
+					struct tm utc;
+					time_t now;
+					time(&now);
+					utc = *gmtime(&now);
+					off = (now - mktime(&utc)) / 60;
+				}
 #endif
-#endif /* !HAVE_TIMEZONE */
+#endif /* !HAVE_VAR_TIMEZONE */
 #endif /* !HAVE_TM_ZONE */
 #endif /* !HAVE_TM_NAME */
 			}
@@ -526,17 +572,17 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 
 #ifdef SYSV_EXT
 		case 'n':	/* same as \n */
-			NEEDS(1);
+			FILL_PADDING(1);
 			*s++ = '\n';
 			continue;
 
 		case 't':	/* same as \t */
-			NEEDS(1);
+			FILL_PADDING(1);
 			*s++ = '\t';
 			continue;
 
 		case 'D':	/* date as %m/%d/%y */
-			STRFTIME("%m/%d/%y", timeptr);
+			STRFTIME("%m/%d/%y");
 			continue;
 
 		case 'e':	/* day of month, blank padded */
@@ -544,15 +590,15 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			continue;
 
 		case 'r':	/* time as %I:%M:%S %p */
-			STRFTIME("%I:%M:%S %p", timeptr);
+			STRFTIME("%I:%M:%S %p");
 			continue;
 
 		case 'R':	/* time as %H:%M */
-			STRFTIME("%H:%M", timeptr);
+			STRFTIME("%H:%M");
 			continue;
 
 		case 'T':	/* time as %H:%M:%S */
-			STRFTIME("%H:%M:%S", timeptr);
+			STRFTIME("%H:%M:%S");
 			continue;
 #endif
 
@@ -581,8 +627,8 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 				     timeptr->tm_year + 1900L);
 			if (w < 0) goto err;
 			for (i = 3; i < 6; i++)
-				if (islower(s[i]))
-					s[i] = toupper(s[i]);
+				if (ISLOWER(s[i]))
+					s[i] = TOUPPER(s[i]);
 			s += w;
 			continue;
 #endif
@@ -609,7 +655,7 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 
 		case 'u':
 		/* ISO 8601: Weekday as a decimal number [1 (Monday) - 7] */
-			FMT('0', 0, "d", timeptr->tm_wday == 0 ? 7 : timeptr->tm_wday);
+			FMT('0', 1, "d", timeptr->tm_wday == 0 ? 7 : timeptr->tm_wday);
 			continue;
 #endif	/* POSIX2_DATE */
 
@@ -634,7 +680,7 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 				y = 1900L + timeptr->tm_year;
 
 			if (*format == 'G')
-				FMT('0', 0, "ld", y);
+				FMT('0', 1, "ld", y);
 			else
 				FMT('0', 2, "ld", y % 100);
 			continue;
@@ -659,8 +705,7 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			{
 				long n = ts->tv_nsec;
 
-				if (precision == 0) continue;
-				if (precision < 0) {
+				if (precision <= 0) {
 				    precision = w;
 				}
 				NEEDS(precision);
@@ -723,29 +768,25 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 
 		default:
 		unknown:
-			tp = sp;
 			i = format - sp + 1;
+			tp = sp;
+			precision = -1;
+			flags = 0;
+			padding = 0;
 			break;
 		}
 		if (i) {
-			if (!(flags & BIT_OF(LEFT)) && precision > i) {
-				NEEDS(precision);
-				memset(s, padding ? padding : ' ', precision - i);
-				s += precision - i;
-			}
-			else {
-				NEEDS(i);
-			}
+			FILL_PADDING(i);
 			memcpy(s, tp, i);
 			switch (flags & (BIT_OF(UPPER)|BIT_OF(LOWER))) {
 			case BIT_OF(UPPER):
 				do {
-					if (islower(*s)) *s = toupper(*s);
+					if (ISLOWER(*s)) *s = TOUPPER(*s);
 				} while (s++, --i);
 				break;
 			case BIT_OF(LOWER):
 				do {
-					if (isupper(*s)) *s = tolower(*s);
+					if (ISUPPER(*s)) *s = TOLOWER(*s);
 				} while (s++, --i);
 				break;
 			default:
@@ -754,7 +795,6 @@ rb_strftime(char *s, size_t maxsize, const char *format, const struct tm *timept
 			}
 		}
 	}
-out:
 	if (s >= endp) {
 		goto err;
 	}
