@@ -42,7 +42,13 @@ class TestFileExhaustive < Test::Unit::TestCase
   def test_path
     file = @file
 
-    assert_equal(file, File.open(file) {|f| f.path})
+    assert_equal(file, File.new(file).path)
+    assert_equal(file, File.path(file))
+    o = Object.new
+    class << o; self; end.class_eval do
+      define_method(:to_path) { file }
+    end
+    assert_equal(file, File.path(o))
   end
 
   def assert_integer(n)
@@ -84,9 +90,9 @@ class TestFileExhaustive < Test::Unit::TestCase
       assert_kind_of(String, fs1.inspect)
     end
     assert_raise(Errno::ENOENT) { File.stat(@nofile) }
-    assert_kind_of(File::Stat, File.open(@file) {|f| f.stat})
+    assert_kind_of(File::Stat, File.new(@file).stat)
     assert_raise(Errno::ENOENT) { File.lstat(@nofile) }
-    assert_kind_of(File::Stat, File.open(@file) {|f| f.lstat})
+    assert_kind_of(File::Stat, File.new(@file).lstat)
   end
 
   def test_directory_p
@@ -152,6 +158,17 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert(!(File.readable_real?(@nofile)))
   end
 
+  def test_world_readable_p
+    return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
+    File.chmod(0006, @file)
+    assert(File.world_readable?(@file))
+    File.chmod(0060, @file)
+    assert(!(File.world_readable?(@file)))
+    File.chmod(0600, @file)
+    assert(!(File.world_readable?(@file)))
+    assert(!(File.world_readable?(@nofile)))
+  end
+
   def test_writable_p
     return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
     File.chmod(0400, @file)
@@ -170,6 +187,17 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert(!(File.writable_real?(@nofile)))
   end
 
+  def test_world_writable_p
+    return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
+    File.chmod(0006, @file)
+    assert(File.world_writable?(@file))
+    File.chmod(0060, @file)
+    assert(!(File.world_writable?(@file)))
+    File.chmod(0600, @file)
+    assert(!(File.world_writable?(@file)))
+    assert(!(File.world_writable?(@nofile)))
+  end
+
   def test_executable_p
     return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
     File.chmod(0100, @file)
@@ -244,7 +272,7 @@ class TestFileExhaustive < Test::Unit::TestCase
 
   def test_atime
     t1 = File.atime(@file)
-    t2 = File.open(@file) {|f| f.atime}
+    t2 = File.new(@file).atime
     assert_kind_of(Time, t1)
     assert_kind_of(Time, t2)
     assert_equal(t1, t2)
@@ -253,7 +281,7 @@ class TestFileExhaustive < Test::Unit::TestCase
 
   def test_mtime
     t1 = File.mtime(@file)
-    t2 = File.open(@file) {|f| f.mtime}
+    t2 = File.new(@file).mtime
     assert_kind_of(Time, t1)
     assert_kind_of(Time, t2)
     assert_equal(t1, t2)
@@ -262,7 +290,7 @@ class TestFileExhaustive < Test::Unit::TestCase
 
   def test_ctime
     t1 = File.ctime(@file)
-    t2 = File.open(@file) {|f| f.ctime}
+    t2 = File.new(@file).ctime
     assert_kind_of(Time, t1)
     assert_kind_of(Time, t2)
     assert_equal(t1, t2)
@@ -273,7 +301,7 @@ class TestFileExhaustive < Test::Unit::TestCase
     return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
     assert_equal(1, File.chmod(0444, @file))
     assert_equal(0444, File.stat(@file).mode % 01000)
-    assert_equal(0, File.open(@file) {|f| f.chmod(0222)})
+    assert_equal(0, File.new(@file).chmod(0222))
     assert_equal(0222, File.stat(@file).mode % 01000)
     File.chmod(0600, @file)
     assert_raise(Errno::ENOENT) { File.chmod(0600, @nofile) }
@@ -384,27 +412,26 @@ class TestFileExhaustive < Test::Unit::TestCase
 
   def test_extname
     assert(".test", File.extname(@file))
-    prefixes = ["", "/", ".", "/.", "bar/.", "/bar/."]
-    infixes = ["", " ", "."]
-    infixes2 = infixes + [".ext "]
-    appendixes = [""]
+    assert_equal("", File.extname("foo"))
+    assert_equal("", File.extname("/foo"))
+    assert_equal("", File.extname(".foo"))
+    assert_equal("", File.extname("/.foo"))
+    assert_equal("", File.extname("bar/.foo"))
+    assert_equal("", File.extname("/bar/.foo"))
+    assert_equal(".ext", File.extname("foo.ext"))
+    assert_equal(".ext", File.extname("/foo.ext"))
+    assert_equal(".ext", File.extname(".foo.ext"))
+    assert_equal(".ext", File.extname("/.foo.ext"))
+    assert_equal(".ext", File.extname("bar/.foo.ext"))
+    assert_equal(".ext", File.extname("/bar/.foo.ext"))
+    assert_equal("", File.extname(""))
     if /cygwin|mingw|mswin|bccwin/ =~ RUBY_PLATFORM
-      appendixes << " " << "." << "::$DATA" << "::$DATA.bar"
+      assert_equal("", File.extname("foo "))
+      assert_equal(".ext", File.extname("foo.ext "))
+      assert_equal(".ext", File.extname("foo.ext."))
+      assert_equal(".ext", File.extname("foo.ext::$DATA"))
+      assert_equal("", File.extname("foo::$DATA.ext"))
     end
-    prefixes.each do |prefix|
-      appendixes.each do |appendix|
-        infixes.each do |infix|
-          path = "#{prefix}foo#{infix}#{appendix}"
-          assert_equal("", File.extname(path), "File.extname(#{path.inspect})")
-        end
-        infixes2.each do |infix|
-          path = "#{prefix}foo#{infix}.ext#{appendix}"
-          assert_equal(".ext", File.extname(path), "File.extname(#{path.inspect})")
-        end
-      end
-    end
-    bug3175 = '[ruby-core:29627]'
-    assert_equal(".rb", File.extname("/tmp//bla.rb"), bug3175)
   end
 
   def test_split
@@ -417,6 +444,9 @@ class TestFileExhaustive < Test::Unit::TestCase
     s = "foo" + File::SEPARATOR + "bar" + File::SEPARATOR + "baz"
     assert_equal(s, File.join("foo", "bar", "baz"))
     assert_equal(s, File.join(["foo", "bar", "baz"]))
+    o = Object.new
+    def o.to_path; "foo"; end
+    assert_equal(s, File.join(o, "bar", "baz"))
     assert_equal(s, File.join("foo" + File::SEPARATOR, "bar", File::SEPARATOR + "baz"))
   end
 
@@ -440,7 +470,7 @@ class TestFileExhaustive < Test::Unit::TestCase
     f.close
     make_file("foo", @file)
 
-    assert_raise(IOError) { File.open(@file) {|f| f.truncate(0)} }
+    assert_raise(IOError) { File.new(@file).truncate(0) }
   rescue NotImplementedError
   end
 
@@ -478,6 +508,7 @@ class TestFileExhaustive < Test::Unit::TestCase
       assert_equal(File.socket?(f), test(?S, f))
       assert_equal(File.setuid?(f), test(?u, f))
       assert_equal(File.writable?(f), test(?w, f))
+      assert_equal(File.world_writable?(f), test(?W, f))
       assert_equal(File.executable?(f), test(?x, f))
       assert_equal(File.executable_real?(f), test(?X, f))
       assert_equal(File.zero?(f), test(?z, f))
@@ -497,6 +528,7 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert_raise(ArgumentError) { test }
     assert_raise(Errno::ENOENT) { test(?A, @nofile) }
     assert_raise(ArgumentError) { test(?a) }
+    assert_raise(ArgumentError) { test("\0".ord) }
   end
 
   def test_stat_init
@@ -593,6 +625,16 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert(File::Stat.new(@file).readable_real?)
   end
 
+  def test_stat_world_readable_p
+    return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
+    File.chmod(0006, @file)
+    assert(File::Stat.new(@file).world_readable?)
+    File.chmod(0060, @file)
+    assert(!(File::Stat.new(@file).world_readable?))
+    File.chmod(0600, @file)
+    assert(!(File::Stat.new(@file).world_readable?))
+  end
+
   def test_stat_writable_p
     return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
     File.chmod(0400, @file)
@@ -609,6 +651,16 @@ class TestFileExhaustive < Test::Unit::TestCase
     assert(File::Stat.new(@file).writable_real?)
   end
 
+  def test_stat_world_writable_p
+    return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
+    File.chmod(0006, @file)
+    assert(File::Stat.new(@file).world_writable?)
+    File.chmod(0060, @file)
+    assert(!(File::Stat.new(@file).world_writable?))
+    File.chmod(0600, @file)
+    assert(!(File::Stat.new(@file).world_writable?))
+  end
+
   def test_stat_executable_p
     return if /cygwin|mswin|bccwin|mingw|emx/ =~ RUBY_PLATFORM
     File.chmod(0100, @file)
