@@ -31,8 +31,8 @@ class TestStringScanner < Test::Unit::TestCase
   def test_s_allocate
     s = StringScanner.allocate
     assert_equal '#<StringScanner (uninitialized)>', s.inspect.sub(/StringScanner_C/, 'StringScanner')
-    assert_raises(UNINIT_ERROR) { s.eos? }
-    assert_raises(UNINIT_ERROR) { s.scan(/a/) }
+    assert_raise(UNINIT_ERROR) { s.eos? }
+    assert_raise(UNINIT_ERROR) { s.scan(/a/) }
     s.string = 'test'
     assert_equal '#<StringScanner 0/4 @ "test">', s.inspect.sub(/StringScanner_C/, 'StringScanner')
     assert_nothing_raised(UNINIT_ERROR) { s.eos? }
@@ -288,15 +288,9 @@ class TestStringScanner < Test::Unit::TestCase
     assert_equal true, s.getch.tainted?
     assert_nil s.getch
 
-    kc_backup = $KCODE
-    begin
-      $KCODE = 'EUC'
-      s = StringScanner.new("\244\242")
-      assert_equal "\244\242", s.getch
-      assert_nil s.getch
-    ensure
-      $KCODE = kc_backup
-    end
+    s = StringScanner.new("\244\242".force_encoding("euc-jp"))
+    assert_equal "\244\242".force_encoding("euc-jp"), s.getch
+    assert_nil s.getch
 
     s = StringScanner.new('test')
     s.scan(/te/)
@@ -322,16 +316,10 @@ class TestStringScanner < Test::Unit::TestCase
     assert_equal true, s.get_byte.tainted?
     assert_nil s.get_byte
 
-    kc_backup = $KCODE
-    begin
-      $KCODE = 'EUC'
-      s = StringScanner.new("\244\242")
-      assert_equal "\244", s.get_byte
-      assert_equal "\242", s.get_byte
-      assert_nil s.get_byte
-    ensure
-      $KCODE = kc_backup
-    end
+    s = StringScanner.new("\244\242".force_encoding("euc-jp"))
+    assert_equal "\244".force_encoding("euc-jp"), s.get_byte
+    assert_equal "\242".force_encoding("euc-jp"), s.get_byte
+    assert_nil s.get_byte
 
     s = StringScanner.new('test')
     s.scan(/te/)
@@ -424,16 +412,9 @@ class TestStringScanner < Test::Unit::TestCase
     assert_nil           s[0]
 
 
-    kc_backup = $KCODE
-    begin
-      $KCODE = 'EUC'
-      s = StringScanner.new("\244\242")
-      s.getch
-      assert_equal "\244\242", s[0]
-    ensure
-      $KCODE = kc_backup
-    end
-
+    s = StringScanner.new("\244\242".force_encoding("euc-jp"))
+    s.getch
+    assert_equal "\244\242".force_encoding("euc-jp"), s[0]
 
     str = 'test'
     str.taint
@@ -555,4 +536,139 @@ class TestStringScanner < Test::Unit::TestCase
     s.terminate
     assert_nil s.matched_size
   end
+
+  def test_encoding
+    ss = StringScanner.new("\xA1\xA2".force_encoding("euc-jp"))
+    assert_equal(Encoding::EUC_JP, ss.scan(/./e).encoding)
+  end
+
+  def test_generic_regexp
+    ss = StringScanner.new("\xA1\xA2".force_encoding("euc-jp"))
+    t = ss.scan(/./)
+    assert_equal("\xa1\xa2".force_encoding("euc-jp"), t)
+  end
+
+  def test_set_pos
+    s = StringScanner.new("test string")
+    s.pos = 7
+    assert_equal("ring", s.rest)
+  end
+
+  def test_match_p
+    s = StringScanner.new("test string")
+    assert_equal(4, s.match?(/\w+/))
+    assert_equal(4, s.match?(/\w+/))
+    assert_equal(nil, s.match?(/\s+/))
+  end
+
+  def test_check
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal("Foo", s.check(/Foo/))
+    assert_equal(0, s.pos)
+    assert_equal("Foo", s.matched)
+    assert_equal(nil, s.check(/Bar/))
+    assert_equal(nil, s.matched)
+  end
+
+  def test_scan_full
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal(4, s.scan_full(/Foo /, false, false))
+    assert_equal(0, s.pos)
+    assert_equal(nil, s.scan_full(/Baz/, false, false))
+    assert_equal("Foo ", s.scan_full(/Foo /, false, true))
+    assert_equal(0, s.pos)
+    assert_equal(nil, s.scan_full(/Baz/, false, false))
+    assert_equal(4, s.scan_full(/Foo /, true, false))
+    assert_equal(4, s.pos)
+    assert_equal(nil, s.scan_full(/Baz /, false, false))
+    assert_equal("Bar ", s.scan_full(/Bar /, true, true))
+    assert_equal(8, s.pos)
+    assert_equal(nil, s.scan_full(/az/, false, false))
+  end
+
+  def test_exist_p
+    s = StringScanner.new("test string")
+    assert_equal(3, s.exist?(/s/))
+    assert_equal(0, s.pos)
+    s.scan(/test/)
+    assert_equal(2, s.exist?(/s/))
+    assert_equal(4, s.pos)
+    assert_equal(nil, s.exist?(/e/))
+  end
+
+  def test_skip_until
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal(3, s.skip_until(/Foo/))
+    assert_equal(3, s.pos)
+    assert_equal(4, s.skip_until(/Bar/))
+    assert_equal(7, s.pos)
+    assert_equal(nil, s.skip_until(/Qux/))
+  end
+
+  def test_check_until
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal("Foo", s.check_until(/Foo/))
+    assert_equal(0, s.pos)
+    assert_equal("Foo Bar", s.check_until(/Bar/))
+    assert_equal(0, s.pos)
+    assert_equal(nil, s.check_until(/Qux/))
+  end
+
+  def test_search_full
+    s = StringScanner.new("Foo Bar Baz")
+    assert_equal(8, s.search_full(/Bar /, false, false))
+    assert_equal(0, s.pos)
+    assert_equal("Foo Bar ", s.search_full(/Bar /, false, true))
+    assert_equal(0, s.pos)
+    assert_equal(8, s.search_full(/Bar /, true, false))
+    assert_equal(8, s.pos)
+    assert_equal("Baz", s.search_full(/az/, true, true))
+    assert_equal(11, s.pos)
+  end
+
+  def test_peek
+    s = StringScanner.new("test string")
+    assert_equal("test st", s.peek(7))
+    assert_equal("test st", s.peek(7))
+    s.scan(/test/)
+    assert_equal(" stri", s.peek(5))
+    assert_equal(" string", s.peek(10))
+    s.scan(/ string/)
+    assert_equal("", s.peek(10))
+  end
+
+  def test_unscan
+    s = StringScanner.new('test string')
+    assert_equal("test", s.scan(/\w+/))
+    s.unscan
+    assert_equal("te", s.scan(/../))
+    assert_equal(nil, s.scan(/\d/))
+    assert_raise(ScanError) { s.unscan }
+  end
+
+  def test_rest
+    s = StringScanner.new('test string')
+    assert_equal("test string", s.rest)
+    s.scan(/test/)
+    assert_equal(" string", s.rest)
+    s.scan(/ string/)
+    assert_equal("", s.rest)
+    s.scan(/ string/)
+  end
+
+  def test_rest_size
+    s = StringScanner.new('test string')
+    assert_equal(11, s.rest_size)
+    s.scan(/test/)
+    assert_equal(7, s.rest_size)
+    s.scan(/ string/)
+    assert_equal(0, s.rest_size)
+    s.scan(/ string/)
+  end
+
+  def test_inspect
+    s = StringScanner.new('test string test')
+    s.scan(/test strin/)
+    assert_equal('#<StringScanner 10/16 "...strin" @ "g tes...">', s.inspect)
+  end
 end
