@@ -2,6 +2,8 @@ require 'optparse'
 require 'erb'
 require 'fileutils'
 
+NUM_ELEM_BYTELOOKUP = 2
+
 C_ESC = {
   "\\" => "\\\\",
   '"' => '\"',
@@ -34,11 +36,11 @@ class StrSet
           seq = $'
           set_result = []
           set.scan(/[^,]+/) {|range|
-            if /\A([0-9a-f][0-9a-f])-([0-9a-f][0-9a-f])\z/ =~ range
+            if /\A([0-9a-f][0-9a-f])-([0-9a-f][0-9a-f])\z/i =~ range
               b = $1.to_i(16)
               e = $2.to_i(16)
               set_result << (b..e)
-            elsif /\A([0-9a-f][0-9a-f])\z/ =~ range
+            elsif /\A([0-9a-f][0-9a-f])\z/i =~ range
               byte = $1.to_i(16)
               set_result << (byte..byte)
             else
@@ -60,7 +62,8 @@ class StrSet
   end
 
   def hash
-    @pat.hash
+    return @hash if defined? @hash
+    @hash = @pat.hash
   end
 
   def eql?(other)
@@ -153,6 +156,33 @@ class StrSet
   end
 end
 
+class ArrayCode
+  def initialize(type, name)
+    @type = type
+    @name = name
+    @len = 0;
+    @content = ''
+  end
+
+  def length
+    @len
+  end
+
+  def insert_at_last(num, str)
+    newnum = self.length + num
+    @content << str
+    @len += num
+  end
+
+  def to_s
+    <<"End"
+static const #{@type}
+#{@name}[#{@len}] = {
+#{@content}};
+End
+  end
+end
+
 class ActionMap
   def self.parse(hash)
     h = {}
@@ -167,11 +197,12 @@ class ActionMap
   end
 
   def hash
+    return @hash if defined? @hash
     hash = 0
     @map.each {|k,v|
       hash ^= k.hash ^ v.hash
     }
-    hash
+    @hash = hash
   end
 
   def eql?(other)
@@ -207,7 +238,7 @@ class ActionMap
         ss.each_firstbyte {|byte, rest|
           h[byte] ||= {}
           if h[byte][rest]
-            raise "ambiguous"
+            raise "ambiguous %s or %s (%02X/%s)" % [h[byte][rest], action, byte, rest]
           end
           h[byte][rest] = action
         }
@@ -236,7 +267,7 @@ class ActionMap
 
   def format_offsets(min, max, offsets)
     offsets = offsets[min..max]
-    code = "{ %d, %d,\n" % [min, max]
+    code = "%d, %d,\n" % [min, max]
     0.step(offsets.length-1,16) {|i|
       code << "    "
       code << offsets[i,8].map {|off| "%3d," % off.to_s }.join('')
@@ -246,10 +277,41 @@ class ActionMap
       end
       code << "\n"
     }
-    code << '}'
     code
   end
 
+  UsedName = {}
+
+  StrMemo = {}
+
+  def str_name(bytes)
+    size = @bytes_code.length
+    rawbytes = [bytes].pack("H*")
+
+    n = nil
+    if !n && !(suf = rawbytes.gsub(/[^A-Za-z0-9_]/, '')).empty? && !UsedName[nn = "str1_" + suf] then n = nn end
+    if !n && !UsedName[nn = "str1_" + bytes] then n = nn end
+    n ||= "str1s_#{size}"
+
+    StrMemo[bytes] = n
+    UsedName[n] = true
+    n
+  end
+
+  def gen_str(bytes)
+    if n = StrMemo[bytes]
+      n
+    else
+      len = bytes.length/2
+      size = @bytes_code.length
+      n = str_name(bytes)
+      @bytes_code.insert_at_last(1 + len,
+        "\#define #{n} makeSTR1(#{size})\n" +
+        "    makeSTR1LEN(#{len})," + bytes.gsub(/../, ' 0x\&,') + "\n\n")
+      n
+    end
+  end
+
   def generate_info(info)
     case info
     when :nomap
@@ -272,10 +334,14 @@ class ActionMap
       "o2(0x#$1,0x#$2)"
     when /\A([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])\z/i
       "o3(0x#$1,0x#$2,0x#$3)"
-    when /\A([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])\z/i
+    when /\A([0-9a-f][0-9a-f])(3[0-9])([0-9a-f][0-9a-f])(3[0-9])\z/i
+      "g4(0x#$1,0x#$2,0x#$3,0x#$4)"
+    when /\A(f[0-7])([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])([0-9a-f][0-9a-f])\z/i
       "o4(0x#$1,0x#$2,0x#$3,0x#$4)"
-    when /\A&/ # pointer to BYTE_LOOKUP structure
-      info.to_s
+    when /\A([0-9a-f][0-9a-f]){4,259}\z/i
+      gen_str(info.upcase)
+    when /\A\/\*BYTE_LOOKUP\*\// # pointer to BYTE_LOOKUP structure
+      $'.to_s
     else
       raise "unexpected action: #{info.inspect}"
     end
@@ -285,7 +351,7 @@ class ActionMap
     infos = infos.map {|info| generate_info(info) }
     maxlen = infos.map {|info| info.length }.max
     columns = maxlen <= 16 ? 4 : 2
-    code = "{\n"
+    code = ""
     0.step(infos.length-1, columns) {|i|
       code << "    "
       is = infos[i,columns]
@@ -294,11 +360,10 @@ class ActionMap
       }
       code << "\n"
     }
-    code << "}"
     code
   end
 
-  def generate_lookup_node(name, table)
+  def generate_lookup_node(bytes_code, words_code, name, table)
     offsets = []
     infos = []
     infomap = {}
@@ -322,44 +387,41 @@ class ActionMap
     offsets_key = [min, max, offsets[min..max]]
     if n = OffsetsMemo[offsets_key]
       offsets_name = n
-      offsets_code = ''
     else
       offsets_name = "#{name}_offsets"
-      offsets_code = <<"End"
-static const unsigned char
-#{offsets_name}[#{2+max-min+1}] = #{format_offsets(min,max,offsets)};
-End
       OffsetsMemo[offsets_key] = offsets_name
+      size = bytes_code.length
+      bytes_code.insert_at_last(2+max-min+1,
+        "\#define #{offsets_name} #{size}\n" +
+        format_offsets(min,max,offsets) + "\n")
     end
 
     if n = InfosMemo[infos]
       infos_name = n
-      infos_code = ''
     else
       infos_name = "#{name}_infos"
-      infos_code = <<"End"
-static const struct byte_lookup* const
-#{infos_name}[#{infos.length}] = #{format_infos(infos)};
-End
       InfosMemo[infos] = infos_name
+
+      size = words_code.length
+      words_code.insert_at_last(infos.length,
+        "\#define #{infos_name} WORDINDEX2INFO(#{size})\n" +
+        format_infos(infos) + "\n")
     end
 
-    r = offsets_code + infos_code + <<"End"
-static const BYTE_LOOKUP
-#{name} = {
+    size = words_code.length
+    words_code.insert_at_last(NUM_ELEM_BYTELOOKUP,
+      "\#define #{name} WORDINDEX2INFO(#{size})\n" +
+      <<"End" + "\n")
     #{offsets_name},
-    #{infos_name}
-};
-
+    #{infos_name},
 End
-    r
   end
 
   PreMemo = {}
   PostMemo = {}
   NextName = "a"
 
-  def generate_node(code, name_hint=nil, valid_encoding=nil)
+  def generate_node(bytes_code, words_code, name_hint=nil, valid_encoding=nil)
     if n = PreMemo[[self,valid_encoding]]
       return n
     end
@@ -371,7 +433,7 @@ End
       else
         name_hint2 = nil
         name_hint2 = "#{name_hint}_#{'%02X' % byte}" if name_hint
-        table[byte] = "&" + rest.generate_node(code, name_hint2, rest_valid_encoding)
+        table[byte] = "/*BYTE_LOOKUP*/" + rest.generate_node(bytes_code, words_code, name_hint2, rest_valid_encoding)
       end
     }
 
@@ -386,9 +448,129 @@ End
 
     PreMemo[[self,valid_encoding]] = PostMemo[table] = name_hint
 
-    code << generate_lookup_node(name_hint, table)
+    generate_lookup_node(bytes_code, words_code, name_hint, table)
     name_hint
   end
+
+  def gennode(bytes_code, words_code, name_hint=nil, valid_encoding=nil)
+    @bytes_code = bytes_code
+    @words_code = words_code
+    name = generate_node(bytes_code, words_code, name_hint, valid_encoding)
+    @bytes_code = nil
+    @words_code = nil
+    return name
+  end
+end
+
+def citrus_mskanji_cstomb(csid, index)
+  case csid
+  when 0
+    index
+  when 1
+    index + 0x80
+  when 2, 3
+    row = index >> 8
+    raise "invalid byte sequence" if row < 0x21
+    if csid == 3
+      if row <= 0x2F
+        offset = (row == 0x22 || row >= 0x26) ? 0xED : 0xF0
+      elsif row >= 0x4D && row <= 0x7E
+        offset = 0xCE
+      else
+        raise "invalid byte sequence"
+      end
+    else
+      raise "invalid byte sequence" if row > 0x97
+      offset = (row < 0x5F) ? 0x81 : 0xC1
+    end
+    col = index & 0xFF
+    raise "invalid byte sequence" if (col < 0x21 || col > 0x7E)
+
+    row -= 0x21
+    col -= 0x21
+    if (row & 1) == 0
+      col += 0x40
+      col += 1 if (col >= 0x7F)
+    else
+      col += 0x9F;
+    end
+    row = row / 2 + offset
+    (row << 8) | col
+  end.to_s(16)
+end
+
+def citrus_euc_cstomb(csid, index)
+  case csid
+  when 0x0000
+    index
+  when 0x8080
+    index | 0x8080
+  when 0x0080
+    index | 0x8E80
+  when 0x8000
+    index | 0x8F8080
+  end.to_s(16)
+end
+
+def citrus_cstomb(ces, csid, index)
+  case ces
+  when 'mskanji'
+    citrus_mskanji_cstomb(csid, index)
+  when 'euc'
+    citrus_euc_cstomb(csid, index)
+  end
+end
+
+SUBDIR = %w/APPLE AST BIG5 CNS CP EBCDIC GB GEORGIAN ISO646 ISO-8859 JIS KAZAKH KOI KS MISC TCVN/
+
+
+def citrus_decode_mapsrc(ces, csid, mapsrcs)
+  table = []
+  mapsrcs.split(',').each do |mapsrc|
+    path = [$srcdir]
+    mode = nil
+    if mapsrc.rindex('UCS', 0)
+      mode = :from_ucs
+      from = mapsrc[4..-1]
+      path << SUBDIR.find{|x| from.rindex(x, 0) }
+    else
+      mode = :to_ucs
+      path << SUBDIR.find{|x| mapsrc.rindex(x, 0) }
+    end
+    path << mapsrc.gsub(':', '@')
+    path = File.join(*path)
+    path << ".src"
+    path[path.rindex('/')] = '%'
+    STDERR.puts 'load mapsrc %s' % path if VERBOSE_MODE
+    open(path) do |f|
+      f.each_line do |l|
+        break if /^BEGIN_MAP/ =~ l
+      end
+      f.each_line do |l|
+        next if /^\s*(?:#|$)/ =~ l
+          break if /^END_MAP/ =~ l
+        case mode
+        when :from_ucs
+          case l
+          when /0x(\w+)\s*-\s*0x(\w+)\s*=\s*INVALID/
+            # Citrus OOB_MODE
+          when /(0x\w+)\s*=\s*(0x\w+)/
+            table.push << [$1.hex, citrus_cstomb(ces, csid, $2.hex)]
+          else
+            raise "unknown notation '%s'"% l
+          end
+        when :to_ucs
+          case l
+          when /(0x\w+)\s*=\s*(0x\w+)/
+            table.push << [citrus_cstomb(ces, csid, $1.hex), $2.hex]
+          else
+            raise "unknown notation '%s'"% l
+          end
+        end
+      end
+    end
+  end
+  return table
 end
 
 def encode_utf8(map)
@@ -406,7 +588,7 @@ def transcode_compile_tree(name, from, map)
   map = encode_utf8(map)
   h = {}
   map.each {|k, v|
-    h[k] = v
+    h[k] = v unless h[k] # use first mapping
   }
   am = ActionMap.parse(h)
 
@@ -418,15 +600,21 @@ def transcode_compile_tree(name, from, map)
     valid_encoding = nil
   end
 
-  code = ''
-  defined_name = am.generate_node(code, name, valid_encoding)
-  return defined_name, code, max_input
+  defined_name = am.gennode(TRANSCODE_GENERATED_BYTES_CODE, TRANSCODE_GENERATED_WORDS_CODE, name, valid_encoding)
+  return defined_name, max_input
 end
 
 TRANSCODERS = []
+TRANSCODE_GENERATED_TRANSCODER_CODE = ''
 
-def transcode_tblgen(from, to, map)
-  STDERR.puts "converter from #{from} to #{to}" if VERBOSE_MODE
+def transcode_tbl_only(from, to, map)
+  if VERBOSE_MODE
+    if from.empty? || to.empty?
+      STDERR.puts "converter for #{from.empty? ? to : from}"
+    else
+      STDERR.puts "converter from #{from} to #{to}"
+    end
+  end
   id_from = from.tr('^0-9A-Za-z', '_')
   id_to = to.tr('^0-9A-Za-z', '_')
   if from == "UTF-8"
@@ -437,7 +625,12 @@ def transcode_tblgen(from, to, map)
     tree_name = "from_#{id_from}_to_#{id_to}"
   end
   map = encode_utf8(map)
-  real_tree_name, tree_code, max_input = transcode_compile_tree(tree_name, from, map)
+  real_tree_name, max_input = transcode_compile_tree(tree_name, from, map)
+  return map, tree_name, real_tree_name, max_input
+end
+
+def transcode_tblgen(from, to, map)
+  map, tree_name, real_tree_name, max_input = transcode_tbl_only(from, to, map)
   transcoder_name = "rb_#{tree_name}"
   TRANSCODERS << transcoder_name
   input_unit_length = UnitLength[from]
@@ -445,23 +638,35 @@ def transcode_tblgen(from, to, map)
   transcoder_code = <<"End"
 static const rb_transcoder
 #{transcoder_name} = {
-    #{c_esc from}, #{c_esc to}, &#{real_tree_name},
+    #{c_esc from}, #{c_esc to}, #{real_tree_name},
+    TRANSCODE_TABLE_INFO,
     #{input_unit_length}, /* input_unit_length */
     #{max_input}, /* max_input */
     #{max_output}, /* max_output */
-    stateless_converter, /* stateful_type */
+    asciicompat_converter, /* asciicompat_type */
+    0, NULL, NULL, /* state_size, state_init, state_fini */
     NULL, NULL, NULL, NULL,
     NULL, NULL, NULL
 };
 End
-  tree_code + "\n" + transcoder_code
+  TRANSCODE_GENERATED_TRANSCODER_CODE << transcoder_code
+  ''
 end
 
 def transcode_generate_node(am, name_hint=nil)
   STDERR.puts "converter for #{name_hint}" if VERBOSE_MODE
-  code = ''
-  am.generate_node(code, name_hint)
-  code
+  name = am.gennode(TRANSCODE_GENERATED_BYTES_CODE, TRANSCODE_GENERATED_WORDS_CODE, name_hint)
+  ''
+end
+
+def transcode_generated_code
+  TRANSCODE_GENERATED_BYTES_CODE.to_s +
+    TRANSCODE_GENERATED_WORDS_CODE.to_s +
+    "\#define TRANSCODE_TABLE_INFO " +
+    "#{OUTPUT_PREFIX}byte_array, #{TRANSCODE_GENERATED_BYTES_CODE.length}, " +
+    "#{OUTPUT_PREFIX}word_array, #{TRANSCODE_GENERATED_WORDS_CODE.length}, " +
+    "sizeof(unsigned int)\n" +
+    TRANSCODE_GENERATED_TRANSCODER_CODE
 end
 
 def transcode_register_code
@@ -528,26 +733,23 @@ ValidEncoding = {
                     {81-fe}{30-39}{81-fe}{30-39}',
 }
 
-{
-  'ASCII-8BIT'  => '1byte',
-  'ISO-8859-1'  => '1byte',
-  'ISO-8859-2'  => '1byte',
-  'ISO-8859-3'  => '1byte',
-  'ISO-8859-4'  => '1byte',
-  'ISO-8859-5'  => '1byte',
-  'ISO-8859-6'  => '1byte',
-  'ISO-8859-7'  => '1byte',
-  'ISO-8859-8'  => '1byte',
-  'ISO-8859-9'  => '1byte',
-  'ISO-8859-10' => '1byte',
-  'ISO-8859-11' => '1byte',
-  'ISO-8859-13' => '1byte',
-  'ISO-8859-14' => '1byte',
-  'ISO-8859-15' => '1byte',
-  'Windows-31J' => 'Shift_JIS',
-}.each {|k, v|
-  ValidEncoding[k] = ValidEncoding.fetch(v)
-}
+def set_valid_byte_pattern (encoding, pattern_or_label)
+  pattern =
+    if ValidEncoding[pattern_or_label]
+      ValidEncoding[pattern_or_label]
+    else
+      pattern_or_label
+    end
+  if ValidEncoding[encoding] and ValidEncoding[encoding]!=pattern
+    raise ArgumentError, "trying to change valid byte pattern for encoding #{encoding} from #{ValidEncoding[encoding]} to #{pattern}"
+  end
+  ValidEncoding[encoding] = pattern
+end
+
+# the following may be used in different places, so keep them here for the moment
+set_valid_byte_pattern 'ASCII-8BIT', '1byte'
+set_valid_byte_pattern 'Windows-31J', 'Shift_JIS'
+set_valid_byte_pattern 'eucJP-ms', 'EUC-JP'
 
 def make_signature(filename, src)
   "src=#{filename.dump}, len=#{src.length}, checksum=#{src.sum}"
@@ -566,9 +768,17 @@ op.parse!
 
 VERBOSE_MODE = verbose_mode
 
+OUTPUT_FILENAME = output_filename
+OUTPUT_PREFIX = output_filename ? File.basename(output_filename)[/\A[A-Za-z0-9_]*/] : ""
+OUTPUT_PREFIX.sub!(/\A_+/, '')
+OUTPUT_PREFIX.sub!(/_*\z/, '_')
+
+TRANSCODE_GENERATED_BYTES_CODE = ArrayCode.new("unsigned char", "#{OUTPUT_PREFIX}byte_array")
+TRANSCODE_GENERATED_WORDS_CODE = ArrayCode.new("unsigned int", "#{OUTPUT_PREFIX}word_array")
+
 arg = ARGV.shift
-dir = File.dirname(arg)
-$:.unshift dir unless $:.include? dir
+$srcdir = File.dirname(arg)
+$:.unshift $srcdir unless $:.include? $srcdir
 src = File.read(arg)
 src.force_encoding("ascii-8bit") if src.respond_to? :force_encoding
 this_script = File.read(__FILE__)
@@ -585,7 +795,7 @@ if !force_mode && output_filename && File.readable?(output_filename)
     if %r{/\* src="([0-9a-z_.-]+)",} =~ line
       name = $1
       next if name == File.basename(arg) || name == File.basename(__FILE__)
-      path = File.join(dir, name)
+      path = File.join($srcdir, name)
       if File.readable? path
         chk_signature << "/* #{make_signature(name, File.read(path))} */\n"
       end
@@ -606,14 +816,16 @@ if VERBOSE_MODE
 end
 
 libs1 = $".dup
-erb_result = ERB.new(src, nil, '%').result(binding)
+erb = ERB.new(src, nil, '%')
+erb.filename = arg
+erb_result = erb.result(binding)
 libs2 = $".dup
 
 libs = libs2 - libs1
 lib_sigs = ''
 libs.each {|lib|
   lib = File.basename(lib)
-  path = File.join(dir, lib)
+  path = File.join($srcdir, lib)
   if File.readable? path
     lib_sigs << "/* #{make_signature(lib, File.read(path))} */\n"
   end
